#
# Copyright 2006-2007, 2013 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.
#
# Backends for the various URL types we support (http, https, ftp, local)

import ftplib
import io
import os
import subprocess
import tempfile
import urllib

import requests

from ..logger import log


#########################
# isoreader abstraction #
#########################


class _XorrisoReader:
    def __init__(self, location):
        self._location = location
        self._cache_file_list = self._make_file_list()

    def _make_file_list(self):
        delim = "VIRTINST_BEGINLIST"
        cmd = ["xorriso", "-indev", self._location, "-print", delim, "-find"]

        log.debug("Generating iso filelist: %s", cmd)
        output = subprocess.check_output(cmd, stderr=subprocess.DEVNULL, universal_newlines=True)
        return output.split(delim, 1)[1].strip().splitlines()

    def grabFile(self, url, scratchdir):
        tmp = tempfile.NamedTemporaryFile(
            prefix="virtinst-iso", suffix="-" + os.path.basename(url), dir=scratchdir
        )

        cmd = ["xorriso", "-osirrox", "on", "-indev", self._location, "-extract", url, tmp.name]
        log.debug("Extracting iso file: %s", cmd)
        subprocess.check_output(cmd, stderr=subprocess.DEVNULL)
        return open(tmp.name, "rb").read()

    def hasFile(self, url):
        return ("'.%s'" % url) in self._cache_file_list


###########################
# Fetcher implementations #
###########################


class _URLFetcher:
    """
    This is a generic base class for fetching/extracting files from
    a media source, such as CD ISO, or HTTP/HTTPS/FTP server
    """

    _block_size = 16384
    _is_iso = False

    def __init__(self, location, scratchdir, meter):
        self.location = location
        self.scratchdir = scratchdir
        self.meter = meter

        log.debug("Using scratchdir=%s", scratchdir)
        self._prepare()

    ####################
    # Internal helpers #
    ####################

    def _make_full_url(self, filename):
        """
        Generate a full fetchable URL from the passed filename, which
        is relative to the self.location
        """
        if not filename:
            return self.location
        return os.path.join(self.location, filename)

    def _grabURL(self, filename, fileobj, fullurl=None):
        """
        Download the filename from self.location, and write contents to
        fileobj
        """
        if fullurl:
            url = fullurl
        else:
            url = self._make_full_url(filename)

        try:
            urlobj, size = self._grabber(url)
        except Exception as e:
            msg = _("Couldn't acquire file %(url)s: %(error)s") % {"url": url, "error": str(e)}
            raise ValueError(msg) from None

        log.debug("Fetching URI: %s", url)
        msg = _("Retrieving '%(filename)s'") % {"filename": os.path.basename(filename)}
        self.meter.start(msg, size)

        self._write(urlobj, fileobj)
        self.meter.end()

    def _write(self, urlobj, fileobj):
        """
        Write the contents of urlobj to python file like object fileobj
        """
        total = 0
        while 1:
            buff = urlobj.read(self._block_size)
            if not buff:
                break
            fileobj.write(buff)
            total += len(buff)
            self.meter.update(total)
        fileobj.flush()
        return total

    def _grabber(self, url):
        """
        Returns the urlobj, size for the passed URL. urlobj is whatever
        data needs to be passed to self._write
        """
        raise NotImplementedError("must be implemented in subclass")

    ##############
    # Public API #
    ##############

    def is_iso(self):
        """
        If this is a fetcher for local CDROM/ISO
        """
        return self._is_iso

    def _prepare(self):
        """
        Perform any necessary setup
        """

    def _cleanup(self):
        """
        Perform any necessary cleanup
        """

    def __del__(self):
        self._cleanup()

    def can_access(self):
        """
        Return True if the location URL seems to be valid
        """
        return True

    def _hasFile(self, url):
        raise NotImplementedError("Must be implemented in subclass")

    def hasFile(self, filename):
        """
        Return True if self.location has the passed filename
        """
        url = self._make_full_url(filename)
        ret = self._hasFile(url)
        log.debug("hasFile(%s) returning %s", url, ret)
        return ret

    def acquireFile(self, filename, fullurl=None):
        """
        Grab the passed filename from self.location and save it to
        a temporary file, returning the temp filename
        """
        fn = None
        try:
            fileobj = tempfile.NamedTemporaryFile(
                prefix="virtinst-",
                suffix="-" + os.path.basename(filename),
                dir=self.scratchdir,
                delete=False,
            )
            fn = fileobj.name

            self._grabURL(filename, fileobj, fullurl=fullurl)
            log.debug("Saved file to %s", fn)
            return fn
        except BaseException:  # pragma: no cover
            if fn and os.path.exists(fn):
                os.unlink(fn)
            raise

    def acquireFileContent(self, filename):
        """
        Grab the passed filename from self.location and return it as a string
        """
        fileobj = io.BytesIO()
        self._grabURL(filename, fileobj)
        return fileobj.getvalue().decode("utf-8")


class _HTTPURLFetcher(_URLFetcher):
    _session = None

    def _prepare(self):
        self._session = requests.Session()

    def _cleanup(self):
        if self._session:
            try:
                self._session.close()
            except Exception:  # pragma: no cover
                log.debug("Error closing requests.session", exc_info=True)
        self._session = None

    def can_access(self):
        return self.hasFile("")

    def _hasFile(self, url):
        """
        We just do a HEAD request to see if the file exists
        """
        try:
            response = self._session.head(url, allow_redirects=True)
            response.raise_for_status()
        except Exception as e:  # pragma: no cover
            log.debug("HTTP hasFile request failed: %s", str(e))
            return False
        return True

    def _grabber(self, url):
        """
        Use requests for this
        """
        response = self._session.get(url, stream=True)
        response.raise_for_status()
        try:
            size = int(response.headers.get("content-length"))
        except Exception:  # pragma: no cover
            size = None
        return response, size

    def _write(self, urlobj, fileobj):
        """
        The requests object doesn't have a file-like read() option, so
        we need to implement it ourselves
        """
        total = 0
        for data in urlobj.iter_content(chunk_size=self._block_size):
            fileobj.write(data)
            total += len(data)
            self.meter.update(total)
        fileobj.flush()
        return total


class _FTPURLFetcher(_URLFetcher):
    _ftp = None

    def _prepare(self):
        if self._ftp:
            return  # pragma: no cover

        try:
            parsed = urllib.parse.urlparse(self.location)
            self._ftp = ftplib.FTP()
            username = urllib.parse.unquote(parsed.username or "")
            password = urllib.parse.unquote(parsed.password or "")
            self._ftp.connect(parsed.hostname, parsed.port or 0)
            self._ftp.login(username, password)
            # Force binary mode
            self._ftp.voidcmd("TYPE I")
        except Exception as e:  # pragma: no cover
            msg = _("Opening URL %(url)s failed: %(error)s") % {
                "url": self.location,
                "error": str(e),
            }
            raise ValueError(msg) from None

    def _grabber(self, url):
        """
        Use urllib and ftplib to grab the file
        """
        request = urllib.request.Request(url)
        urlobj = urllib.request.urlopen(request)
        size = self._ftp.size(urllib.parse.urlparse(url)[2])
        return urlobj, size

    def _cleanup(self):
        if not self._ftp:
            return  # pragma: no cover

        try:
            self._ftp.quit()
        except Exception:  # pragma: no cover
            log.debug("Error quitting ftp connection", exc_info=True)

        self._ftp = None

    def _hasFile(self, url):
        path = urllib.parse.urlparse(url)[2]

        try:
            try:
                # If it's a file
                self._ftp.size(path)
            except ftplib.all_errors:  # pragma: no cover
                # If it's a dir
                self._ftp.cwd(path)
        except ftplib.all_errors as e:  # pragma: no cover
            log.debug("FTP hasFile: couldn't access %s: %s", url, str(e))
            return False

        return True


class _LocalURLFetcher(_URLFetcher):
    """
    For grabbing files from a local directory
    """

    def _hasFile(self, url):
        parsed = urllib.parse.urlparse(url)
        return os.path.exists(parsed.path)

    def _grabber(self, url):
        parsed = urllib.parse.urlparse(url)
        urlobj = open(parsed.path, "rb")
        size = os.path.getsize(parsed.path)
        return urlobj, size


class _ISOURLFetcher(_URLFetcher):
    _isoreader = None
    _is_iso = True

    def _make_full_url(self, filename):
        return os.path.join("/", filename)

    def _get_isoreader(self):
        if not self._isoreader:
            self._isoreader = _XorrisoReader(self.location)
        return self._isoreader

    def _grabber(self, url):
        if not self._hasFile(url):
            raise RuntimeError("iso doesn't have file=%s" % url)

        output = self._get_isoreader().grabFile(url, self.scratchdir)
        return io.BytesIO(output), len(output)

    def _hasFile(self, url):
        return self._get_isoreader().hasFile(url)


class DirectFetcher(_URLFetcher):
    def _make_full_url(self, filename):
        return filename

    def acquireFile(self, filename, fullurl=None):
        if not fullurl:
            fullurl = filename
        filename = os.path.basename(filename)
        fetcher = fetcherForURI(fullurl, self.scratchdir, self.meter, direct=True)
        return fetcher.acquireFile(filename, fullurl)  # pylint: disable=protected-access

    def _hasFile(self, url):
        return True

    def _grabber(self, url):
        raise RuntimeError("DirectFetcher shouldn't be used for file access.")  # pragma: no cover


def fetcherForURI(uri, scratchdir, meter, direct=False):
    if uri.startswith("http://") or uri.startswith("https://"):
        fclass = _HTTPURLFetcher
    elif uri.startswith("ftp://"):
        fclass = _FTPURLFetcher
    elif direct or os.path.isdir(uri):
        # Pointing to a local tree
        fclass = _LocalURLFetcher
    else:
        # Pointing to a path (e.g. iso), or a block device (e.g. /dev/cdrom)
        fclass = _ISOURLFetcher
    return fclass(uri, scratchdir, meter)



#
# Copyright 2010, 2013 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from ..xmlbuilder import XMLBuilder, XMLProperty, XMLChildProperty


class _InitArg(XMLBuilder):
    XML_NAME = "initarg"
    val = XMLProperty(".")


class _InitEnv(XMLBuilder):
    XML_NAME = "initenv"
    name = XMLProperty("./@name")
    value = XMLProperty(".")


class _BootDevice(XMLBuilder):
    XML_NAME = "boot"
    dev = XMLProperty("./@dev")


class _FirmwareFeature(XMLBuilder):
    XML_NAME = "feature"
    _XML_PROP_ORDER = ["enabled", "name"]

    enabled = XMLProperty("./@enabled", is_yesno=True)
    name = XMLProperty("./@name")


class DomainOs(XMLBuilder):
    """
    Class for generating boot device related XML
    """

    BOOT_DEVICE_HARDDISK = "hd"
    BOOT_DEVICE_CDROM = "cdrom"
    BOOT_DEVICE_FLOPPY = "fd"
    BOOT_DEVICE_NETWORK = "network"
    BOOT_DEVICES = [
        BOOT_DEVICE_HARDDISK,
        BOOT_DEVICE_CDROM,
        BOOT_DEVICE_FLOPPY,
        BOOT_DEVICE_NETWORK,
    ]

    def is_hvm(self):
        return self.os_type == "hvm"

    def is_xenpv(self):
        return self.os_type in ["xen", "linux", "xenpvh"]

    def is_container(self):
        return self.os_type == "exe"

    def is_x86(self):
        return self.arch == "x86_64" or self.arch == "i686"

    def is_q35(self):
        return self.is_x86() and self.machine and "q35" in self.machine

    def is_arm32(self):
        return self.arch == "armv7l"

    def is_arm64(self):
        return self.arch == "aarch64"

    def is_arm(self):
        return self.is_arm32() or self.is_arm64()

    def is_arm_machvirt(self):
        return self.is_arm() and str(self.machine).startswith("virt")

    def is_ppc64(self):
        return self.arch == "ppc64" or self.arch == "ppc64le"

    def is_pseries(self):
        return self.is_ppc64() and str(self.machine).startswith("pseries")

    def is_s390x(self):
        return self.arch == "s390x"

    def is_riscv(self):
        return self.arch == "riscv64" or self.arch == "riscv32"

    def is_riscv_virt(self):
        return self.is_riscv() and str(self.machine).startswith("virt")

    def is_loongarch64(self):
        return self.arch == "loongarch64"

    ##################
    # XML properties #
    ##################

    XML_NAME = "os"
    _XML_PROP_ORDER = [
        "firmware",
        "os_type",
        "arch",
        "machine",
        "firmware_features",
        "loader",
        "loader_ro",
        "loader_secure",
        "loader_type",
        "loader_stateless",
        "nvram",
        "nvram_template",
        "init",
        "initargs",
        "initenvs",
        "initdir",
        "inituser",
        "initgroup",
        "kernel",
        "initrd",
        "kernel_args",
        "dtb",
        "acpi_tb",
        "acpi_tb_type",
        "bootdevs",
        "bootmenu_enable",
        "bootmenu_timeout",
        "bios_useserial",
        "bios_rebootTimeout",
        "smbios_mode",
    ]

    # Shared/Generic boot options
    os_type = XMLProperty("./type")
    arch = XMLProperty("./type/@arch")
    machine = XMLProperty("./type/@machine")
    loader = XMLProperty("./loader")
    loader_ro = XMLProperty("./loader/@readonly", is_yesno=True)
    loader_type = XMLProperty("./loader/@type")
    loader_secure = XMLProperty("./loader/@secure", is_yesno=True)
    loader_stateless = XMLProperty("./loader/@stateless", is_yesno=True)

    # BIOS bootloader options
    def _get_bootorder(self):
        return [dev.dev for dev in self.bootdevs]

    def _set_bootorder(self, newdevs):
        for dev in self.bootdevs:
            self.remove_child(dev)

        for d in newdevs:
            dev = self.bootdevs.add_new()
            dev.dev = d

    bootorder = property(_get_bootorder, _set_bootorder)
    bootdevs = XMLChildProperty(_BootDevice)
    firmware = XMLProperty("./@firmware")
    firmware_features = XMLChildProperty(_FirmwareFeature, relative_xpath="./firmware")
    nvram = XMLProperty("./nvram", do_abspath=True)
    nvram_template = XMLProperty("./nvram/@template")
    bootmenu_enable = XMLProperty("./bootmenu/@enable", is_yesno=True)
    bootmenu_timeout = XMLProperty("./bootmenu/@timeout", is_int=True)
    bios_useserial = XMLProperty("./bios/@useserial", is_yesno=True)
    bios_rebootTimeout = XMLProperty("./bios/@rebootTimeout", is_int=True)
    smbios_mode = XMLProperty("./smbios/@mode")
    nvram_templateFormat = XMLProperty("./nvram/@templateFormat")

    # Host bootloader options
    # Since the elements for a host bootloader are actually directly under
    # <domain> rather than <domain><os>, they are handled via callbacks in
    # the CLI. This is just a placeholder to remind of that fact.

    # Direct kernel boot options
    kernel = XMLProperty("./kernel", do_abspath=True)
    initrd = XMLProperty("./initrd", do_abspath=True)
    kernel_args = XMLProperty("./cmdline")
    dtb = XMLProperty("./dtb", do_abspath=True)
    acpi_tb = XMLProperty("./acpi/table", do_abspath=True)
    acpi_tb_type = XMLProperty("./acpi/table/@type")

    # Container boot options
    init = XMLProperty("./init")
    initargs = XMLChildProperty(_InitArg)
    initenvs = XMLChildProperty(_InitEnv)
    initdir = XMLProperty("./initdir")
    inituser = XMLProperty("./inituser")
    initgroup = XMLProperty("./initgroup")

    def set_initargs_string(self, argstring):
        import shlex

        for obj in self.initargs:
            self.remove_child(obj)
        for val in shlex.split(argstring):
            obj = self.initargs.add_new()
            obj.val = val

    ##################
    # Default config #
    ##################

    def set_defaults(self, guest):
        if self.is_container() and not self.init:
            if guest.is_full_os_container():
                self.init = "/sbin/init"
            else:
                self.init = "/bin/sh"



#
# Copyright 2006-2009, 2013, 2014 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os
import shutil
import subprocess
import tempfile

from ..logger import log


def _run_initrd_commands(initrd, tempdir):
    log.debug("Appending to the initrd.")

    find_proc = subprocess.Popen(
        ["find", ".", "-print0"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=tempdir
    )
    cpio_proc = subprocess.Popen(
        ["cpio", "--create", "--null", "--quiet", "--format=newc", "--owner=0:0"],
        stdin=find_proc.stdout,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        cwd=tempdir,
    )
    f = open(initrd, "ab")
    gzip_proc = subprocess.Popen(["gzip"], stdin=cpio_proc.stdout, stdout=f, stderr=subprocess.PIPE)
    cpio_proc.wait()
    find_proc.wait()
    gzip_proc.wait()
    f.close()

    finderr = find_proc.stderr.read()
    cpioerr = cpio_proc.stderr.read()
    gziperr = gzip_proc.stderr.read()
    if finderr:  # pragma: no cover
        log.debug("find stderr=%s", finderr)
    if cpioerr:  # pragma: no cover
        log.debug("cpio stderr=%s", cpioerr)
    if gziperr:  # pragma: no cover
        log.debug("gzip stderr=%s", gziperr)

    if (
        cpio_proc.returncode != 0 or find_proc.returncode != 0 or gzip_proc.returncode != 0
    ):  # pragma: no cover
        raise RuntimeError("Failed to inject files into initrd")


def _run_iso_commands(iso, tempdir, cloudinit=False):
    # These three programs all behave similarly for our needs, and
    # different distros only have some available. xorriso is apparently
    # the actively maintained variant that should be available everywhere
    # and without any license issues. Some more info here:
    # https://wiki.debian.org/genisoimage
    programs = ["xorrisofs", "genisoimage", "mkisofs"]
    for program in programs:
        if shutil.which(program):
            break

    cmd = [program, "-o", iso, "-J", "-input-charset", "utf8", "-rational-rock"]
    if cloudinit:
        cmd.extend(["-V", "cidata"])
    cmd.append(tempdir)
    log.debug("Running iso build command: %s", cmd)
    output = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
    log.debug("cmd output: %s", output)


def _perform_generic_injections(injections, scratchdir, media, cb, **kwargs):
    if not injections:
        return

    tempdir = tempfile.mkdtemp(dir=scratchdir)
    try:
        os.chmod(tempdir, 0o775)

        for filename in injections:
            if type(filename) is tuple:
                filename, dst = filename
            else:
                dst = os.path.basename(filename)

            log.debug("Injecting src=%s dst=%s into media=%s", filename, dst, media)
            shutil.copy(filename, os.path.join(tempdir, dst))

        return cb(media, tempdir, **kwargs)
    finally:
        shutil.rmtree(tempdir)


def perform_initrd_injections(initrd, injections, scratchdir):
    """
    Insert files into the root directory of the initial ram disk
    """
    _perform_generic_injections(injections, scratchdir, initrd, _run_initrd_commands)


def perform_cdrom_injections(injections, scratchdir, cloudinit=False):
    """
    Insert files into the root directory of a generated cdrom
    """
    if cloudinit:
        iso_suffix = "-cloudinit.iso"
    else:
        iso_suffix = "-unattended.iso"
    fileobj = tempfile.NamedTemporaryFile(
        prefix="virtinst-", suffix=iso_suffix, dir=scratchdir, delete=False
    )
    iso = fileobj.name

    try:
        _perform_generic_injections(
            injections, scratchdir, iso, _run_iso_commands, cloudinit=cloudinit
        )
    except Exception:  # pragma: no cover
        os.unlink(iso)
        raise

    return iso



# Copyright (C) 2009, 2013, 2014 Red Hat, Inc.
# Copyright (C) 2009 Cole Robinson <crobinso@redhat.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from gi.repository import Gtk

from virtinst import log

from .asyncjob import vmmAsyncJob


####################################################################
# Build toolbar shutdown button menu (manager and details toolbar) #
####################################################################


class _VMMenu(Gtk.Menu):
    def __init__(self, src, current_vm_cb, show_open=True):
        Gtk.Menu.__init__(self)
        self._parent = src
        self._current_vm_cb = current_vm_cb
        self._show_open = show_open

        self._init_state()

    def _add_action(self, label, widgetname, cb):
        item = Gtk.MenuItem.new_with_mnemonic(label)

        item.vmm_widget_name = widgetname
        if cb:

            def _cb(_menuitem):
                _vm = self._current_vm_cb()
                if _vm:
                    return cb(self._parent, _vm)

            item.connect("activate", _cb)

        self.add(item)
        return item

    def _init_state(self):
        raise NotImplementedError()

    def update_widget_states(self, vm):
        raise NotImplementedError()


class VMShutdownMenu(_VMMenu):
    """
    Shutdown submenu for reboot, forceoff, reset, etc.
    """

    def _init_state(self):
        self._add_action(_("_Reboot"), "reboot", VMActionUI.reboot)
        self._add_action(_("_Shut Down"), "shutdown", VMActionUI.shutdown)
        self._add_action(_("F_orce Reset"), "reset", VMActionUI.reset)
        self._add_action(_("_Force Off"), "destroy", VMActionUI.destroy)
        self.add(Gtk.SeparatorMenuItem())
        self._add_action(_("Sa_ve"), "save", VMActionUI.save)

        self.get_accessible().set_name("vmm-shutdown-menu")
        self.show_all()

    def update_widget_states(self, vm):
        statemap = {
            "reboot": bool(vm and vm.is_stoppable()),
            "shutdown": bool(vm and vm.is_stoppable()),
            "reset": bool(vm and vm.is_stoppable()),
            "destroy": bool(vm and vm.is_destroyable()),
            "save": bool(vm and vm.is_destroyable()),
        }

        for child in self.get_children():
            name = getattr(child, "vmm_widget_name", None)
            if name in statemap:
                child.set_sensitive(statemap[name])


class VMActionMenu(_VMMenu):
    """
    VM submenu for run, pause, shutdown, clone, etc
    """

    def _init_state(self):
        self._add_action(_("_Run"), "run", VMActionUI.run)
        self._add_action(_("_Pause"), "suspend", VMActionUI.suspend)
        self._add_action(_("R_esume"), "resume", VMActionUI.resume)
        s = self._add_action(_("_Shut Down"), "shutdown", None)
        s.set_submenu(VMShutdownMenu(self._parent, self._current_vm_cb))

        self.add(Gtk.SeparatorMenuItem())
        self._add_action(_("Clone..."), "clone", VMActionUI.clone)
        self._add_action(_("Migrate..."), "migrate", VMActionUI.migrate)
        self._add_action(_("_Delete"), "delete", VMActionUI.delete)

        if self._show_open:
            self.add(Gtk.SeparatorMenuItem())
            self._add_action(_("_Open"), "show", VMActionUI.show)

        self.get_accessible().set_name("vm-action-menu")
        self.show_all()

    def update_widget_states(self, vm):
        statemap = {
            "run": bool(vm and vm.is_runable()),
            "shutdown": bool(vm and vm.is_stoppable()),
            "suspend": bool(vm and vm.is_stoppable()),
            "resume": bool(vm and vm.is_paused()),
            "migrate": bool(vm and vm.is_stoppable()),
            "clone": bool(vm and vm.is_cloneable()),
        }
        vismap = {
            "suspend": bool(vm and not vm.is_paused()),
            "resume": bool(vm and vm.is_paused()),
        }

        for child in self.get_children():
            name = getattr(child, "vmm_widget_name", None)
            if child.get_submenu():
                child.get_submenu().update_widget_states(vm)
            if name in statemap:
                child.set_sensitive(statemap[name])
            if name in vismap:
                child.set_visible(vismap[name])

    def change_run_text(self, text):
        for child in self.get_children():
            if getattr(child, "vmm_widget_name", None) == "run":
                child.get_child().set_label(text)


class VMActionUI:
    """
    Singleton object for handling VM actions, asking for confirmation,
    showing errors/progress dialogs, etc.
    """

    @staticmethod
    def save_cancel(asyncjob, vm):
        log.debug("Cancelling save job")
        if not vm:
            return  # pragma: no cover

        try:
            vm.abort_job()
        except Exception as e:
            log.exception("Error cancelling save job")
            asyncjob.show_warning(_("Error cancelling save job: %s") % str(e))
            return

        asyncjob.job_canceled = True  # pragma: no cover

    @staticmethod
    def save(src, vm):
        if not src.err.chkbox_helper(
            src.config.get_confirm_poweroff,
            src.config.set_confirm_poweroff,
            text1=_("Are you sure you want to save '%s'?") % vm.get_name(),
        ):
            return

        _cancel_cb = None
        if vm.supports_domain_job_info():
            _cancel_cb = (VMActionUI.save_cancel, vm)

        def cb(asyncjob):
            vm.save(meter=asyncjob.get_meter())

        def finish_cb(error, details):
            if error is not None:  # pragma: no cover
                error = _("Error saving domain: %s") % error
                src.err.show_err(error, details=details)

        progWin = vmmAsyncJob(
            cb,
            [],
            finish_cb,
            [],
            _("Saving Virtual Machine"),
            _("Saving virtual machine memory to disk "),
            src.topwin,
            cancel_cb=_cancel_cb,
        )
        progWin.run()

    @staticmethod
    def destroy(src, vm):
        if not src.err.chkbox_helper(
            src.config.get_confirm_forcepoweroff,
            src.config.set_confirm_forcepoweroff,
            text1=_("Are you sure you want to force poweroff '%s'?") % vm.get_name(),
            text2=_(
                "This will immediately poweroff the VM without "
                "shutting down the OS and may cause data loss."
            ),
        ):
            return

        log.debug("Destroying vm '%s'", vm.get_name())
        vmmAsyncJob.simple_async_noshow(vm.destroy, [], src, _("Error shutting down domain"))

    @staticmethod
    def suspend(src, vm):
        if not src.err.chkbox_helper(
            src.config.get_confirm_pause,
            src.config.set_confirm_pause,
            text1=_("Are you sure you want to pause '%s'?") % vm.get_name(),
        ):
            return

        log.debug("Pausing vm '%s'", vm.get_name())
        vmmAsyncJob.simple_async_noshow(vm.suspend, [], src, _("Error pausing domain"))

    @staticmethod
    def resume(src, vm):
        log.debug("Unpausing vm '%s'", vm.get_name())
        vmmAsyncJob.simple_async_noshow(vm.resume, [], src, _("Error unpausing domain"))

    @staticmethod
    def run(src, vm):
        log.debug("Starting vm '%s'", vm.get_name())

        if vm.has_managed_save():

            def errorcb(error, details):
                # This is run from the main thread
                res = src.err.show_err(
                    _("Error restoring domain: %s") % error,
                    details=details,
                    text2=_(
                        "The domain could not be restored. Would you like\n"
                        "to remove the saved state and perform a regular\n"
                        "start up?"
                    ),
                    dialog_type=Gtk.MessageType.WARNING,
                    buttons=Gtk.ButtonsType.YES_NO,
                    modal=True,
                )

                if not res:
                    return

                try:
                    vm.remove_saved_image()
                    VMActionUI.run(src, vm)
                except Exception as e:  # pragma: no cover
                    src.err.show_err(_("Error removing domain state: %s") % str(e))

            # VM will be restored, which can take some time, so show progress
            title = _("Restoring Virtual Machine")
            text = _("Restoring virtual machine memory from disk")
            vmmAsyncJob.simple_async(vm.startup, [], src, title, text, "", errorcb=errorcb)

        else:
            # Regular startup
            errorintro = _("Error starting domain")
            vmmAsyncJob.simple_async_noshow(vm.startup, [], src, errorintro)

    @staticmethod
    def shutdown(src, vm):
        if not src.err.chkbox_helper(
            src.config.get_confirm_poweroff,
            src.config.set_confirm_poweroff,
            text1=_("Are you sure you want to poweroff '%s'?") % vm.get_name(),
        ):
            return

        log.debug("Shutting down vm '%s'", vm.get_name())
        vmmAsyncJob.simple_async_noshow(vm.shutdown, [], src, _("Error shutting down domain"))

    @staticmethod
    def reboot(src, vm):
        if not src.err.chkbox_helper(
            src.config.get_confirm_poweroff,
            src.config.set_confirm_poweroff,
            text1=_("Are you sure you want to reboot '%s'?") % vm.get_name(),
        ):
            return

        log.debug("Rebooting vm '%s'", vm.get_name())
        vmmAsyncJob.simple_async_noshow(vm.reboot, [], src, _("Error rebooting domain"))

    @staticmethod
    def reset(src, vm):
        if not src.err.chkbox_helper(
            src.config.get_confirm_forcepoweroff,
            src.config.set_confirm_forcepoweroff,
            text1=_("Are you sure you want to force reset '%s'?") % vm.get_name(),
            text2=_(
                "This will immediately reset the VM without "
                "shutting down the OS and may cause data loss."
            ),
        ):
            return

        log.debug("Resetting vm '%s'", vm.get_name())
        vmmAsyncJob.simple_async_noshow(vm.reset, [], src, _("Error resetting domain"))

    @staticmethod
    def delete(src, vm):
        from .delete import vmmDeleteDialog

        vmmDeleteDialog.show_instance(src, vm)

    @staticmethod
    def migrate(src, vm):
        from .migrate import vmmMigrateDialog

        vmmMigrateDialog.show_instance(src, vm)

    @staticmethod
    def clone(src, vm):
        from .clone import vmmCloneVM

        vmmCloneVM.show_instance(src, vm)

    @staticmethod
    def show(src, vm):
        from .vmwindow import vmmVMWindow

        vmmVMWindow.get_instance(src, vm).show()



# Copyright (C) 2013-2014 Red Hat, Inc.
# Copyright (C) 2013 Cole Robinson <crobinso@redhat.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import datetime
import glob
import io
import os

from gi.repository import GdkPixbuf
from gi.repository import Gtk
from gi.repository import Pango

from virtinst import DomainSnapshot
from virtinst import generatename
from virtinst import log
from virtinst import xmlutil

from ..lib import uiutil
from ..asyncjob import vmmAsyncJob
from ..baseclass import vmmGObjectUI


mimemap = {
    "image/x-portable-pixmap": "ppm",
    "image/png": "png",
}


def _make_screenshot_pixbuf(mime, sdata):
    loader = GdkPixbuf.PixbufLoader.new_with_mime_type(mime)
    loader.write(sdata)
    pixbuf = loader.get_pixbuf()
    loader.close()

    maxsize = 450

    def _scale(big, small, maxsize):
        if big <= maxsize:
            return big, small  # pragma: no cover
        factor = float(maxsize) / float(big)
        return maxsize, int(factor * float(small))

    width = pixbuf.get_width()
    height = pixbuf.get_height()
    if width > height:
        width, height = _scale(width, height, maxsize)
    else:
        height, width = _scale(height, width, maxsize)  # pragma: no cover

    return pixbuf.scale_simple(width, height, GdkPixbuf.InterpType.BILINEAR)


def _mime_to_ext(val, reverse=False):
    for m, e in mimemap.items():
        if val == m and not reverse:
            return e
        if val == e and reverse:
            return m
    log.debug(
        "Don't know how to convert %s=%s to %s",  # pragma: no cover
        reverse and "extension" or "mime",
        val,
        reverse and "mime" or "extension",
    )


class vmmSnapshotNew(vmmGObjectUI):
    __gsignals__ = {
        "snapshot-created": (vmmGObjectUI.RUN_FIRST, None, [str]),
    }

    def __init__(self, vm):
        vmmGObjectUI.__init__(self, "snapshotsnew.ui", "snapshot-new")
        self.vm = vm

        self._init_ui()

        self.builder.connect_signals(
            {
                "on_snapshot_new_delete_event": self.close,
                "on_snapshot_new_cancel_clicked": self.close,
                "on_snapshot_new_name_changed": self._name_changed_cb,
                "on_snapshot_new_name_activate": self._ok_clicked_cb,
                "on_snapshot_new_ok_clicked": self._ok_clicked_cb,
                "on_snapshot_new_mode_toggled": self._mode_toggled_cb,
                "on_snapshot_new_memory_toggled": self._memory_toggled_cb,
            }
        )
        self.bind_escape_key_close()

    #######################
    # Standard UI methods #
    #######################

    def show(self, parent):
        log.debug("Showing new snapshot wizard")
        self._reset_state()
        self.topwin.resize(1, 1)
        self.topwin.set_transient_for(parent)
        self.topwin.present()

    def close(self, ignore1=None, ignore2=None):
        log.debug("Closing new snapshot wizard")
        self.topwin.hide()
        return 1

    def _cleanup(self):
        self.vm = None

    ###########
    # UI init #
    ###########

    def _init_snapshot_mode(self):
        mode_external = self.widget("snapshot-new-mode-external")

        capsinfo = self.vm.xmlobj.lookup_capsinfo()

        if not capsinfo.guest.supports_externalSnapshot():
            mode_external.set_sensitive(False)
            mode_external.set_tooltip_text(
                _("external snapshots not supported with this libvirt connection")
            )

    def _init_memory_path(self):
        mempaths = self.widget("snapshot-new-memory-path")

        model = Gtk.ListStore(str)
        mempaths.set_model(model)
        uiutil.init_combo_text_column(mempaths, 0)

    def _init_ui(self):
        buf = Gtk.TextBuffer()
        self.widget("snapshot-new-description").set_buffer(buf)
        self._init_snapshot_mode()
        self._init_memory_path()

    def _reset_snapshot_mode(self):
        mode_external = self.widget("snapshot-new-mode-external")
        mode_internal = self.widget("snapshot-new-mode-internal")

        use_external = mode_external.is_sensitive()

        if use_external:
            current_mode = self._get_current_mode()
            if current_mode == "internal":
                use_external = False

        if use_external:
            mode_external.set_active(True)
        else:
            mode_internal.set_active(True)

    def _reset_snapshot_memory_path(self):
        mode_box = self.widget("snapshot-new-mode-external")
        self._set_memory_path_visibility(mode_box)

        self.widget("snapshot-new-memory-auto").set_active(True)

    def _reset_state(self):
        basename = "snapshot"

        def cb(n):
            return generatename.check_libvirt_collision(
                self.vm.get_backend().snapshotLookupByName, n
            )

        default_name = generatename.generate_name(basename, cb, sep="", start_num=1, force_num=True)

        self.widget("snapshot-new-name").set_text(default_name)
        self.widget("snapshot-new-name").emit("changed")
        self.widget("snapshot-new-description").get_buffer().set_text("")
        self.widget("snapshot-new-ok").grab_focus()
        self.widget("snapshot-new-status-text").set_text(self.vm.run_status())
        self.widget("snapshot-new-status-icon").set_from_icon_name(
            self.vm.run_status_icon_name(), Gtk.IconSize.BUTTON
        )

        self._reset_snapshot_mode()

        self._reset_snapshot_memory_path()

        self.widget("snapshot-new-name").grab_focus()

    ############
    # UI utils #
    ############

    def _set_memory_path_visibility(self, mode_box):
        uiutil.set_grid_row_visible(
            self.widget("snapshot-new-memory-label"), mode_box.get_active() and self.vm.is_active()
        )

    def _populate_memory_path(self):
        mempaths = self.widget("snapshot-new-memory-path")

        if mempaths.is_sensitive():
            return

        model = mempaths.get_model()
        paths = []

        model.clear()

        name = self.widget("snapshot-new-name").get_text()
        memname = f"{self.vm.get_name()}-mem.{name}"

        for disk in self.vm.get_xmlobj().devices.disk:
            diskpath = disk.get_source_path()

            if not diskpath:
                continue

            newpath = os.path.join(os.path.dirname(diskpath), memname)

            if newpath not in paths:
                paths.append(newpath)

        for path in paths:
            model.append([path])

        mempaths.set_active(0)

    ###################
    # Create handling #
    ###################

    def _take_screenshot(self):
        stream = None
        try:
            stream = self.vm.conn.get_backend().newStream(0)
            screen = 0
            flags = 0
            mime = self.vm.get_backend().screenshot(stream, screen, flags)

            ret = io.BytesIO()

            def _write_cb(_stream, data, userdata):
                ignore = stream
                ignore = userdata
                ret.write(data)

            stream.recvAll(_write_cb, None)
            return mime, ret.getvalue()
        finally:
            try:
                if stream:
                    stream.finish()
            except Exception:  # pragma: no cover
                pass

    def _get_screenshot(self):
        if not self.vm.is_active():
            log.debug("Skipping screenshot since VM is not active")
            return
        if not self.vm.xmlobj.devices.graphics:
            log.debug("Skipping screenshot since VM has no graphics")
            return

        try:
            # Perform two screenshots, because qemu + qxl has a bug where
            # screenshot generally only shows the data from the previous
            # screenshot request:
            # https://bugs.launchpad.net/qemu/+bug/1314293
            self._take_screenshot()
            mime, sdata = self._take_screenshot()
        except Exception:  # pragma: no cover
            log.exception("Error taking screenshot")
            return

        ext = _mime_to_ext(mime)
        if not ext:
            return  # pragma: no cover

        return mime, sdata

    def _get_mode(self):
        mode_external = self.widget("snapshot-new-mode-external")
        if mode_external.get_active():
            return "external"

        return "internal"

    def _new_finish_cb(self, error, details, newname):
        self.reset_finish_cursor()

        if error is not None:
            error = _("Error creating snapshot: %s") % error
            self.err.show_err(error, details=details)
            return

        self.emit("snapshot-created", newname)
        self.close()

    def _validate_new_snapshot(self):
        name = self.widget("snapshot-new-name").get_text()
        desc = self.widget("snapshot-new-description").get_buffer().get_property("text")
        mode = self._get_mode()

        try:
            newsnap = DomainSnapshot(self.vm.conn.get_backend())
            newsnap.name = name
            newsnap.description = desc or None
            if mode == "external" and self.vm.is_active():
                mempath = uiutil.get_list_selection(self.widget("snapshot-new-memory-path"))
                newsnap.memory_type = mode
                newsnap.memory_file = mempath
            newsnap.get_xml()
            newsnap.validate_generic_name(_("Snapshot"), newsnap.name)
            return newsnap
        except Exception as e:
            return self.err.val_err(_("Error validating snapshot: %s") % e)

    def _do_create_snapshot(self, asyncjob, xml, name, mime, sndata, diskOnly):
        ignore = asyncjob

        self.vm.create_snapshot(xml, diskOnly=diskOnly)

        try:
            cachedir = self.vm.get_cache_dir()
            basesn = os.path.join(cachedir, "snap-screenshot-%s" % name)

            # Remove any pre-existing screenshots so we don't show stale data
            for ext in list(mimemap.values()):
                p = basesn + "." + ext
                if os.path.exists(basesn + "." + ext):
                    os.unlink(p)

            if not mime or not sndata:
                return

            filename = basesn + "." + _mime_to_ext(mime)
            log.debug("Writing screenshot to %s", filename)
            open(filename, "wb").write(sndata)
        except Exception:  # pragma: no cover
            log.exception("Error saving screenshot")

    def _create_new_snapshot(self):
        snap = self._validate_new_snapshot()
        if not snap:
            return

        xml = snap.get_xml()
        name = snap.name
        mime, sndata = self._get_screenshot() or (None, None)
        diskOnly = not self.vm.is_active() and self._get_mode() == "external"

        self.set_finish_cursor()
        progWin = vmmAsyncJob(
            self._do_create_snapshot,
            [xml, name, mime, sndata, diskOnly],
            self._new_finish_cb,
            [name],
            _("Creating snapshot"),
            _("Creating virtual machine snapshot"),
            self.topwin,
        )
        progWin.run()

    def _get_current_mode(self):
        current = self.vm.get_current_snapshot()

        if current is None:
            return None

        if current.is_external():
            return "external"

        return "internal"

    ################
    # UI listeners #
    ################

    def _name_changed_cb(self, src):
        self.widget("snapshot-new-ok").set_sensitive(bool(src.get_text()))
        self._populate_memory_path()

    def _ok_clicked_cb(self, src):
        current_mode = self._get_current_mode()

        if current_mode and current_mode != self._get_mode():
            result = self.err.yes_no(
                _(
                    "Mixing external and internal snapshots for "
                    "the same VM is not recommended. Are you "
                    "sure you want to continue?"
                )
            )

            if not result:
                return

        return self._create_new_snapshot()

    def _mode_toggled_cb(self, src):
        self._set_memory_path_visibility(src)

    def _memory_toggled_cb(self, src):
        mempaths = self.widget("snapshot-new-memory-path")

        if src.get_active():
            mempaths.set_sensitive(False)
        else:
            mempaths.set_sensitive(True)

        self._populate_memory_path()


class vmmSnapshotPage(vmmGObjectUI):
    def __init__(self, vm, builder, topwin):
        vmmGObjectUI.__init__(self, "snapshots.ui", None, builder=builder, topwin=topwin)

        self.vm = vm

        self._initial_populate = False
        self._unapplied_changes = False
        self._snapshot_new = None

        self._snapmenu = None
        self._init_ui()

        self.builder.connect_signals(
            {
                "on_snapshot_add_clicked": self._on_add_clicked,
                "on_snapshot_delete_clicked": self._on_delete_clicked,
                "on_snapshot_start_clicked": self._on_start_clicked,
                "on_snapshot_apply_clicked": self._on_apply_clicked,
                "on_snapshot_list_changed": self._snapshot_selected,
                "on_snapshot_list_button_press_event": self._popup_snapshot_menu,
                "on_snapshot_refresh_clicked": self._on_refresh_clicked,
                "on_snapshot_list_row_activated": self._on_start_clicked,
            }
        )

        self.top_box = self.widget("snapshot-top-box")
        self.widget("snapshot-top-window").remove(self.top_box)
        selection = self.widget("snapshot-list").get_selection()
        selection.emit("changed")
        selection.set_mode(Gtk.SelectionMode.MULTIPLE)
        selection.set_select_function(self._confirm_changes, None)

    ##############
    # Init stuff #
    ##############

    def _cleanup(self):
        self.vm = None
        self._snapmenu = None

        if self._snapshot_new:
            self._snapshot_new.cleanup()
            self._snapshot_new = None

    def _init_ui(self):
        # pylint: disable=redefined-variable-type
        self.widget("snapshot-notebook").set_show_tabs(False)

        buf = Gtk.TextBuffer()
        buf.connect("changed", self._description_changed)
        self.widget("snapshot-description").set_buffer(buf)

        # [name, row label, tooltip, icon name, sortname, current]
        model = Gtk.ListStore(str, str, str, str, str, bool)
        model.set_sort_column_id(4, Gtk.SortType.ASCENDING)

        col = Gtk.TreeViewColumn("")
        col.set_min_width(150)
        col.set_spacing(6)

        img = Gtk.CellRendererPixbuf()
        img.set_property("stock-size", Gtk.IconSize.LARGE_TOOLBAR)
        col.pack_start(img, False)
        col.add_attribute(img, "icon-name", 3)

        txt = Gtk.CellRendererText()
        txt.set_property("ellipsize", Pango.EllipsizeMode.END)
        col.pack_start(txt, False)
        col.add_attribute(txt, "markup", 1)

        img = Gtk.CellRendererPixbuf()
        img.set_property("stock-size", Gtk.IconSize.MENU)
        img.set_property("icon-name", "emblem-default")
        img.set_property("xalign", 0.0)
        col.pack_start(img, False)
        col.add_attribute(img, "visible", 5)

        def _sep_cb(_model, _iter, ignore):
            return not bool(_model[_iter][0])

        slist = self.widget("snapshot-list")
        slist.set_model(model)
        slist.set_tooltip_column(2)
        slist.append_column(col)
        slist.set_row_separator_func(_sep_cb, None)

        # Snapshot popup menu
        menu = Gtk.Menu()

        item = Gtk.MenuItem.new_with_mnemonic(_("_Start snapshot"))
        item.show()
        item.connect("activate", self._on_start_clicked)
        menu.add(item)

        item = Gtk.MenuItem.new_with_mnemonic(_("_Delete snapshot"))
        item.show()
        item.connect("activate", self._on_delete_clicked)
        menu.add(item)

        self._snapmenu = menu

    ###################
    # Functional bits #
    ###################

    def _get_selected_snapshots(self):
        selection = self.widget("snapshot-list").get_selection()

        def add_snap(treemodel, path, it, snaps):
            ignore = path
            try:
                name = treemodel[it][0]
                for snap in self.vm.list_snapshots():
                    if name == snap.get_name():
                        snaps.append(snap)
            except Exception:  # pragma: no cover
                pass

        snaps = []
        selection.selected_foreach(add_snap, snaps)
        return snaps

    def _refresh_snapshots(self, select_name=None):
        self.vm.refresh_snapshots()
        self._populate_snapshot_list(select_name)

    def vmwindow_refresh_vm_state(self):
        if not self._initial_populate:
            self._populate_snapshot_list()

    def _set_error_page(self, msg):
        self._set_snapshot_state(None)
        self.widget("snapshot-notebook").set_current_page(1)
        self.widget("snapshot-error-label").set_text(msg)

    def _populate_snapshot_list(self, select_name=None):
        cursnaps = []
        for i in self._get_selected_snapshots():
            cursnaps.append(i.get_name())

        model = self.widget("snapshot-list").get_model()
        model.clear()

        try:
            snapshots = self.vm.list_snapshots()
        except Exception as e:  # pragma: no cover
            log.exception(e)
            self._set_error_page(_("Error refreshing snapshot list: %s") % str(e))
            return

        has_external = False
        has_internal = False
        for snap in snapshots:
            desc = snap.get_xmlobj().description
            name = snap.get_name()
            state = snap.run_status()
            if snap.is_external():
                has_external = True
                sortname = "3%s" % name
                label = _("%(vm)s\n<span size='small'>VM State: %(state)s (External)</span>")
            else:
                has_internal = True
                sortname = "1%s" % name
                label = _("%(vm)s\n<span size='small'>VM State: %(state)s</span>")

            label = label % {"vm": xmlutil.xml_escape(name), "state": xmlutil.xml_escape(state)}
            model.append(
                [name, label, desc, snap.run_status_icon_name(), sortname, snap.is_current()]
            )

        if has_internal and has_external:
            model.append([None, None, None, None, "2", False])

        def check_selection(treemodel, path, it, snaps):
            if select_name:
                if treemodel[it][0] == select_name:
                    selection.select_path(path)
            elif treemodel[it][0] in snaps:
                selection.select_path(path)

        selection = self.widget("snapshot-list").get_selection()
        model = self.widget("snapshot-list").get_model()
        selection.unselect_all()
        model.foreach(check_selection, cursnaps)

        self._initial_populate = True

    def _read_screenshot_file(self, name):
        if not name:
            return

        cache_dir = self.vm.get_cache_dir()
        basename = os.path.join(cache_dir, "snap-screenshot-%s" % name)
        files = glob.glob(basename + ".*")
        if not files:
            return

        filename = files[0]
        mime = _mime_to_ext(os.path.splitext(filename)[1][1:], reverse=True)
        if not mime:
            return  # pragma: no cover
        return _make_screenshot_pixbuf(mime, open(filename, "rb").read())

    def _set_snapshot_state(self, snap=None):
        self.widget("snapshot-notebook").set_current_page(0)

        xmlobj = snap and snap.get_xmlobj() or None
        name = snap and xmlobj.name or ""
        desc = snap and xmlobj.description or ""
        state = snap and snap.run_status() or ""
        icon = snap and snap.run_status_icon_name() or None
        is_external = snap and snap.is_external() or False
        is_current = snap and snap.is_current() or False

        timestamp = ""
        if snap:
            timestamp = str(datetime.datetime.fromtimestamp(xmlobj.creationTime))

        title = ""
        if name:
            title = _("<b>Snapshot '%(name)s':</b>") % {"name": xmlutil.xml_escape(name)}

        uiutil.set_grid_row_visible(self.widget("snapshot-is-current"), is_current)
        self.widget("snapshot-title").set_markup(title)
        self.widget("snapshot-timestamp").set_text(timestamp)
        self.widget("snapshot-description").get_buffer().set_text(desc)

        self.widget("snapshot-status-text").set_text(state)
        if icon:
            self.widget("snapshot-status-icon").set_from_icon_name(icon, Gtk.IconSize.BUTTON)

        uiutil.set_grid_row_visible(self.widget("snapshot-mode"), is_external)
        if is_external:
            is_mem = xmlobj.memory_type == "external"
            is_disk = [d.snapshot == "external" for d in xmlobj.disks]
            if is_mem and is_disk:
                mode = _("External disk and memory")
            elif is_mem:
                mode = _("External memory only")
            else:
                mode = _("External disk only")
            self.widget("snapshot-mode").set_text(mode)

        sn = self._read_screenshot_file(name)
        self.widget("snapshot-screenshot").set_visible(bool(sn))
        self.widget("snapshot-screenshot-label").set_visible(not bool(sn))
        if sn:
            self.widget("snapshot-screenshot").set_from_pixbuf(sn)

        self.widget("snapshot-add").set_sensitive(True)
        self.widget("snapshot-delete").set_sensitive(bool(snap))
        self.widget("snapshot-start").set_sensitive(bool(snap))
        self.widget("snapshot-apply").set_sensitive(False)
        self._unapplied_changes = False

    def _confirm_changes(self, sel, model, path, path_selected, user_data):
        ignore1 = sel
        ignore2 = path
        ignore3 = model
        ignore4 = user_data

        if not self._unapplied_changes or not path_selected:
            return True

        if self.err.confirm_unapplied_changes():
            self._apply()

        return True

    def _apply(self):
        snaps = self._get_selected_snapshots()
        if not snaps or len(snaps) > 1:
            return False  # pragma: no cover

        snap = snaps[0]
        desc_widget = self.widget("snapshot-description")
        desc = desc_widget.get_buffer().get_property("text") or ""

        xmlobj = snap.get_xmlobj()
        origxml = xmlobj.get_xml()
        xmlobj.description = desc
        newxml = xmlobj.get_xml()

        self.vm.log_redefine_xml_diff(snap, origxml, newxml)
        if newxml == origxml:
            return True  # pragma: no cover

        self.vm.create_snapshot(newxml, redefine=True)
        snap.ensure_latest_xml()
        return True

    #############
    # Listeners #
    #############

    def _popup_snapshot_menu(self, src, event):
        ignore = src
        if event.button != 3:
            return
        self._snapmenu.popup_at_pointer(event)

    def close(self, ignore1=None, ignore2=None):
        if self._snapshot_new:
            self._snapshot_new.close()
        return 1

    def _description_changed(self, ignore):
        snaps = self._get_selected_snapshots()
        desc_widget = self.widget("snapshot-description")
        desc = desc_widget.get_buffer().get_property("text") or ""

        if len(snaps) == 1 and snaps[0].get_xmlobj().description != desc:
            self._unapplied_changes = True

        self.widget("snapshot-apply").set_sensitive(True)

    def _on_apply_clicked(self, ignore):
        self._apply()
        self._refresh_snapshots()

    def _snapshot_created_cb(self, src, newname):
        self._refresh_snapshots(newname)

    def _on_add_clicked(self, ignore):
        if not self._snapshot_new:
            self._snapshot_new = vmmSnapshotNew(self.vm)
            self._snapshot_new.connect("snapshot-created", self._snapshot_created_cb)
        if self.vm.has_managed_save():
            result = self.err.ok_cancel(
                _("Saved memory state will not be part of the snapshot"),
                _(
                    "The domain is currently saved. Due to technical "
                    "limitations that saved memory state will not become part "
                    "of the snapshot. Running it later will be the same as "
                    "having forced the system off mid-flight. It is "
                    "recommended to snapshot either the running or shut down "
                    "system instead."
                ),
            )
            if not result:
                return
        self._snapshot_new.show(self.topwin)

    def _on_refresh_clicked(self, ignore):
        self._refresh_snapshots()

    def _on_start_clicked(self, ignore, ignore2=None, ignore3=None):
        snaps = self._get_selected_snapshots()
        if not snaps or len(snaps) > 1:
            return  # pragma: no cover

        snap = snaps[0]

        if self.vm.is_active():
            msg = _(
                "Are you sure you want to run the snapshot '%(name)s'? "
                "All the disk changes since the last snapshot was created "
                "will be discarded."
            )
        else:
            msg = _(
                "Are you sure you want to run the snapshot '%(name)s'? "
                "All the disk and configuration changes since the last "
                "snapshot was created will be discarded."
            )
        msg = msg % {"name": snap.get_name()}

        result = self.err.yes_no(msg)
        if not result:
            return

        if self.vm.has_managed_save() and not snap.has_run_state():
            result = self.err.ok_cancel(
                _("Saved state will be removed to avoid filesystem corruption"),
                _(
                    "Snapshot '%s' contains only disk and no memory state. "
                    "Restoring the snapshot would leave the existing saved state "
                    "in place, effectively switching a disk underneath a running "
                    "system. Running the domain afterwards would likely result in "
                    "extensive filesystem corruption. Therefore the saved state "
                    "will be removed before restoring the snapshot."
                )
                % snap.get_name(),
            )
            if not result:
                return
            self.vm.remove_saved_image()

        log.debug("Running snapshot '%s'", snap.get_name())
        vmmAsyncJob.simple_async(
            self.vm.revert_to_snapshot,
            [snap],
            self,
            _("Running snapshot"),
            _("Running snapshot '%s'") % snap.get_name(),
            _("Error running snapshot '%s'") % snap.get_name(),
            finish_cb=self._refresh_snapshots,
        )

    def _on_delete_clicked(self, ignore):
        snaps = self._get_selected_snapshots()
        if not snaps:
            return  # pragma: no cover

        result = self.err.yes_no(
            _("Are you sure you want to permanently delete the selected snapshots?")
        )
        if not result:
            return

        for snap in snaps:
            log.debug("Deleting snapshot '%s'", snap.get_name())
            vmmAsyncJob.simple_async(
                snap.delete,
                [],
                self,
                _("Deleting snapshot"),
                _("Deleting snapshot '%s'") % snap.get_name(),
                _("Error deleting snapshot '%s'") % snap.get_name(),
                finish_cb=self._refresh_snapshots,
            )

    def _snapshot_selected(self, selection):
        ignore = selection
        snap = self._get_selected_snapshots()
        if not snap:
            self._set_error_page(_("No snapshot selected."))
            return
        if len(snap) > 1:
            self._set_error_page(_("Multiple snapshots selected."))
            self.widget("snapshot-start").set_sensitive(False)
            self.widget("snapshot-apply").set_sensitive(False)
            self.widget("snapshot-delete").set_sensitive(True)
            return

        try:
            self._set_snapshot_state(snap[0])
        except Exception as e:  # pragma: no cover
            log.exception(e)
            self._set_error_page(_("Error selecting snapshot: %s") % str(e))



# Copyright (C) 2009, 2013 Red Hat, Inc.
# Copyright (C) 2009 Cole Robinson <crobinso@redhat.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import collections

from gi.repository import Gtk
from gi.repository import Pango

import virtinst
from virtinst import Cloner
from virtinst import log
from virtinst import xmlutil

from .lib import uiutil
from .baseclass import vmmGObjectUI
from .asyncjob import vmmAsyncJob
from .storagebrowse import vmmStorageBrowser


def _get_cloneable_msg(diskinfo):
    """Is the passed path even clone-able"""
    if diskinfo.get_cloneable_msg():
        return diskinfo.get_cloneable_msg()
    if diskinfo.disk.is_empty():
        return _("No storage to clone.")


class _StorageInfo:
    """
    Our wrapper class that is close to Cloner _DiskInfo content.
    We track all user choices explicitly in this class, and then
    serialize them into the actual cloner diskinfo once the
    user clicks 'Clone' to complete the process.
    """

    def __init__(self, vm, cloner, diskinfo):
        self._diskinfo = diskinfo
        self._orig_disk = diskinfo.disk
        self._orig_vm_name = vm.get_name()
        self._new_vm_name = cloner.new_guest.name

        self.share_msg = diskinfo.get_share_msg()
        self.cloneable_msg = _get_cloneable_msg(diskinfo)

        self._manual_path = None
        self._generated_path = diskinfo.new_disk and diskinfo.new_disk.get_source_path()
        if not self._generated_path:
            self._reset_generated_path()

        self._is_clone_requested = diskinfo.is_clone_requested()

        if self.share_msg or not self.is_cloneable():
            self._is_clone_requested = False

    def is_cloneable(self):
        return not bool(self.cloneable_msg)

    def is_clone_requested(self):
        return self._is_clone_requested

    def is_share_requested(self):
        return not self._is_clone_requested

    def warn_about_sharing(self):
        return not self.share_msg

    def get_target(self):
        return self._orig_disk.target

    def get_orig_disk_path(self):
        return self._orig_disk.get_source_path()

    def get_new_disk_path(self):
        if self._manual_path:
            return self._manual_path
        return self._generated_path

    def set_clone_requested(self, val):
        self._is_clone_requested = bool(val)

    def set_manual_path(self, val):
        self._manual_path = val

    def set_new_vm_name(self, val):
        if not val or val == self._new_vm_name:
            return
        self._new_vm_name = val
        self._reset_generated_path()

    def _reset_generated_path(self):
        self._generated_path = Cloner.generate_clone_disk_path(
            self._orig_disk.conn,
            self._orig_vm_name,
            self._new_vm_name,
            self._orig_disk.get_source_path(),
        )

    def set_values_on_diskinfo(self, diskinfo):
        if not self._is_clone_requested:
            diskinfo.set_share_requested()
            diskinfo.new_disk = None
            return

        diskinfo.set_clone_requested()
        sparse = True
        newpath = self.get_new_disk_path()
        diskinfo.set_new_path(newpath, sparse)

    ###################
    # UI info helpers #
    ###################

    def get_tooltip(self):
        lines = []
        lines.append(_("Disk target: %s") % self.get_target())
        lines.append(_("Original path: %s") % self.get_orig_disk_path())
        if self.get_new_disk_path():
            lines.append(_("New path: %s") % self.get_new_disk_path())
        lines.append("\n")

        if self.share_msg:
            lines.append(_("Storage is safe to share: %(reason)s") % {"reason": self.share_msg})
        else:
            lines.append(_("Sharing this storage is potentially dangerous."))

        if self.cloneable_msg:
            lines.append(_("Storage is not cloneable: %(reason)s") % {"reason": self.cloneable_msg})
        return "\n".join(lines)

    def get_markup(self, vm):
        lines = []

        line = ""
        path = self.get_orig_disk_path()
        if path:
            line += path
        else:
            line += _("No storage.")
        lines.append(line)

        line = ""
        if self.is_share_requested():
            line += _("Share disk with %s") % vm.get_name()
        if self.is_clone_requested():
            line += _("Clone this disk")
            sizelabel = self.get_size_label()
            if sizelabel:
                line += " (%s)" % sizelabel
        if line:
            lines.append(line)

        label = "\n".join(lines)
        markup = "<small>%s</small>" % xmlutil.xml_escape(label)
        return markup

    def get_icon_name(self):
        if self._orig_disk.is_floppy():
            return "media-floppy"
        if self._orig_disk.is_cdrom():
            return "media-optical"
        return "drive-harddisk"

    def get_size_label(self):
        size = self._orig_disk.get_size()
        if not size:
            return ""  # pragma: no cover
        return "%.1f GiB" % float(size)


class vmmCloneVM(vmmGObjectUI):
    @classmethod
    def show_instance(cls, parentobj, vm):
        try:
            # Maintain one dialog per connection
            uri = vm.conn.get_uri()
            if cls._instances is None:
                cls._instances = {}
            if uri not in cls._instances:
                cls._instances[uri] = vmmCloneVM()
            cls._instances[uri].show(parentobj.topwin, vm)
        except Exception as e:  # pragma: no cover
            parentobj.err.show_err(_("Error launching clone dialog: %s") % str(e))

    def __init__(self):
        vmmGObjectUI.__init__(self, "clone.ui", "vmm-clone")
        self.vm = None

        self._storage_list = None
        self._storage_browser = None

        self._storage_dialog = self.widget("vmm-change-storage")
        self._storage_dialog.set_transient_for(self.topwin)

        self.builder.connect_signals(
            {
                "on_clone_delete_event": self._close_cb,
                "on_clone_cancel_clicked": self._close_cb,
                "on_clone_ok_clicked": self._finish_clicked_cb,
                "on_storage_selection_changed": self._storage_selection_changed_cb,
                "on_storage_details_clicked": self._storage_details_clicked_cb,
                # Storage subdialog signals
                "on_vmm_change_storage_delete_event": self._storage_dialog_close_cb,
                "on_change_storage_cancel_clicked": self._storage_dialog_close_cb,
                "on_change_storage_ok_clicked": self._storage_dialog_finish_cb,
                "on_change_storage_doclone_toggled": self._storage_dialog_doclone_toggled_cb,
                "on_change_storage_browse_clicked": self._storage_dialog_browse_cb,
            }
        )
        self.bind_escape_key_close()
        self._cleanup_on_app_close()

        self._init_ui()

    #######################
    # Standard UI methods #
    #######################

    @property
    def conn(self):
        return self.vm and self.vm.conn or None

    def show(self, parent, vm):
        log.debug("Showing clone wizard")
        self._set_vm(vm)
        self._reset_state()
        self.topwin.set_transient_for(parent)
        self.topwin.resize(1, 1)
        self.topwin.present()

    def _storage_dialog_close(self):
        self._storage_dialog.hide()
        return 1

    def close(self, ignore1=None, ignore2=None):
        log.debug("Closing clone wizard")
        self._storage_dialog_close()
        self.topwin.hide()

        self._set_vm(None)
        self._storage_list = None
        return 1

    def _vm_removed_cb(self, _conn, vm):
        if self.vm == vm:
            self.close()

    def _set_vm(self, newvm):
        oldvm = self.vm
        if oldvm:
            oldvm.conn.disconnect_by_obj(self)
        if newvm:
            newvm.conn.connect("vm-removed", self._vm_removed_cb)
        self.vm = newvm

    def _cleanup(self):
        self._storage_dialog.destroy()
        self._storage_dialog = None

        if self._storage_browser:
            self._storage_browser.cleanup()
            self._storage_browser = None

    ###########
    # UI init #
    ###########

    def _init_ui(self):
        storage_list = self.widget("storage-list")

        # [disk target, tooltip]
        model = Gtk.ListStore(str, str)
        storage_list.set_model(model)
        storage_list.set_tooltip_column(1)

        cloneCol = Gtk.TreeViewColumn(_("Clone"))
        pathCol = Gtk.TreeViewColumn(_("Storage"))
        pathCol.set_sizing(Gtk.TreeViewColumnSizing.AUTOSIZE)
        pathCol.set_expand(True)

        storage_list.append_column(cloneCol)
        storage_list.append_column(pathCol)

        def separator_cb(_model, _iter):
            return _model[_iter][0] == "separator"

        storage_list.set_row_separator_func(separator_cb)

        chkbox = Gtk.CellRendererToggle()
        chkbox.connect("toggled", self._storage_clone_toggled_cb)
        chkimg = Gtk.CellRendererPixbuf()
        chkimg.set_property("stock-size", Gtk.IconSize.MENU)
        cloneCol.pack_start(chkimg, False)
        cloneCol.pack_start(chkbox, False)

        def chk_cell_data_cb(column, cell, model, _iter, data):
            target = model[_iter][0]
            if target == "separator":
                return
            sinfo = self._storage_list[target]
            visible = sinfo.is_cloneable()
            active = sinfo.is_clone_requested()
            _chkimg = column.get_cells()[0]
            _chkbox = column.get_cells()[1]
            _chkbox.set_property("active", active)
            _chkbox.set_property("visible", visible)
            _chkimg.set_property("visible", not visible)
            icon = "dialog-information"
            if sinfo.warn_about_sharing():
                icon = "dialog-warning"
            _chkimg.set_property("icon-name", icon)
            tooltip = sinfo.get_tooltip()
            if tooltip != model[_iter][1]:
                model[_iter][1] = tooltip

        cloneCol.set_cell_data_func(chkbox, chk_cell_data_cb)
        cloneCol.set_cell_data_func(chkimg, chk_cell_data_cb)

        pathtxt = Gtk.CellRendererText()
        pathCol.set_sort_column_id(0)
        pathtxt.set_property("width-chars", 30)
        pathtxt.set_property("ellipsize", Pango.EllipsizeMode.MIDDLE)
        pathimg = Gtk.CellRendererPixbuf()
        pathimg.set_property("stock-size", Gtk.IconSize.MENU)
        pathimg.set_padding(3, 0)
        pathCol.pack_start(pathimg, False)
        pathCol.pack_start(pathtxt, True)

        def path_cb(column, cell, model, _iter, data):
            target = model[_iter][0]
            if target == "separator":
                return
            sinfo = self._storage_list[target]
            _pathimg = column.get_cells()[0]
            _pathtxt = column.get_cells()[1]
            markup = sinfo.get_markup(self.vm)
            _pathtxt.set_property("markup", markup)
            _pathimg.set_property("icon-name", sinfo.get_icon_name())

        pathCol.set_cell_data_func(pathtxt, path_cb)
        pathCol.set_cell_data_func(pathimg, path_cb)

    def _reset_state(self):
        self.widget("clone-cancel").grab_focus()
        self.widget("clone-new-name").set_text("")

        # Populate default clone values
        cloner = self._build_cloner()
        cloner.prepare()
        self.widget("clone-orig-name").set_text(cloner.src_name)
        self.widget("clone-new-name").set_text(cloner.new_guest.name)

        uiutil.set_grid_row_visible(self.widget("clone-dest-host"), self.conn.is_remote())
        self.widget("clone-dest-host").set_text(self.conn.get_pretty_desc())

        # Collect info about all the VMs disks
        self._storage_list = collections.OrderedDict()

        for diskinfo in cloner.get_diskinfos():
            sinfo = _StorageInfo(self.vm, cloner, diskinfo)
            self._storage_list[sinfo.get_target()] = sinfo

        self._populate_storage_ui()

    #######################
    # Functional routines #
    #######################

    def _build_cloner(self):
        conn = self.conn.get_backend()
        orig_name = self.vm.get_name()
        new_name = self.widget("clone-new-name").get_text()
        cloner = Cloner(conn, src_name=orig_name)
        if new_name:
            cloner.set_clone_name(new_name)
        return cloner

    #######################
    # Storage UI building #
    #######################

    def _set_paths_from_clone_name(self):
        newname = self.widget("clone-new-name").get_text()
        for sinfo in list(self._storage_list.values()):
            sinfo.set_new_vm_name(newname)

    def _populate_storage_ui(self):
        """
        Fill in the storage UI from the collected StorageInfo classes
        """
        model = self.widget("storage-list").get_model()
        model.clear()
        for sinfo in self._storage_list.values():
            model.append([sinfo.get_target(), sinfo.get_tooltip()])
            model.append(["separator", None])

    def _storage_clone_toggled_cb(self, src, treepath):
        model = self.widget("storage-list").get_model()
        row = model[treepath]
        target = row[0]
        sinfo = self._storage_list[target]
        active = not src.get_property("active")
        sinfo.set_clone_requested(active)
        model.emit("row-changed", row.path, row.iter)

    ###########################
    # Storage window handling #
    ###########################

    def _show_storage_window(self):
        tgt = uiutil.get_list_selection(self.widget("storage-list"))
        sinfo = self._storage_list[tgt]

        # If storage paths are dependent on manually entered clone name,
        # make sure they are up to date
        self._set_paths_from_clone_name()

        orig = sinfo.get_orig_disk_path()
        new = sinfo.get_new_disk_path()
        size = sinfo.get_size_label()
        can_clone = sinfo.is_cloneable()
        do_clone = sinfo.is_clone_requested()

        self.widget("change-storage-doclone").set_active(True)
        self.widget("change-storage-doclone").toggled()
        self.widget("change-storage-orig").set_text(orig)
        self.widget("change-storage-target").set_text(tgt)
        self.widget("change-storage-size").set_text(size or "-")
        self.widget("change-storage-doclone").set_active(do_clone)

        self.widget("change-storage-new").set_text(new or "")
        self.widget("change-storage-doclone").set_sensitive(can_clone)

        self.widget("vmm-change-storage").show_all()

    def _storage_dialog_finish(self):
        target = self.widget("change-storage-target").get_text()
        sinfo = self._storage_list[target]

        # Sync 'do clone' checkbox, and main dialog combo
        do_clone = self.widget("change-storage-doclone").get_active()
        sinfo.set_clone_requested(do_clone)

        if not do_clone:
            self._storage_dialog_close()
            return

        new_path = self.widget("change-storage-new").get_text()

        if virtinst.DeviceDisk.path_definitely_exists(self.vm.conn.get_backend(), new_path):
            text1 = _("Cloning will overwrite the existing file")
            text2 = _(
                "Using an existing image will overwrite "
                "the path during the clone process. Are "
                "you sure you want to use this path?"
            )
            res = self.err.yes_no(text1, text2)
            if not res:
                return

        sinfo.set_manual_path(new_path)
        self._storage_dialog_close()

    ###################
    # Finish routines #
    ###################

    def _validate(self, cloner):
        new_paths = []
        warn_str = ""
        for sinfo in self._storage_list.values():
            if sinfo.is_clone_requested():
                new_paths.append(sinfo.get_new_disk_path())
                continue
            if not sinfo.warn_about_sharing():
                continue
            warn_str += "%s: %s\n" % (sinfo.get_target(), sinfo.get_orig_disk_path())

        if warn_str:
            res = self.err.ok_cancel(
                _("Sharing storage may cause data to be overwritten."),
                _(
                    "The following disk devices will be shared with %(vmname)s:"
                    "\n\n%(pathlist)s\n"
                    "Running the new guest could overwrite data in these "
                    "disk images."
                )
                % {"vmname": self.vm.get_name(), "pathlist": warn_str},
            )
            if not res:
                return False

        for diskinfo in cloner.get_diskinfos():
            diskinfo.raise_error()

    def _finish_cb(self, error, details, conn, cloner):
        self.reset_finish_cursor()

        if error is not None:
            error = _("Error creating virtual machine clone '%(vm)s': %(error)s") % {
                "vm": cloner.new_guest.name,
                "error": error,
            }
            self.err.show_err(error, details=details)
            return

        conn.schedule_priority_tick(pollvm=True)
        self.close()

    def _async_clone(self, asyncjob, cloner):
        meter = asyncjob.get_meter()

        refresh_pools = []
        for diskinfo in cloner.get_nonshare_diskinfos():
            disk = diskinfo.new_disk
            pool = disk.get_parent_pool()
            if not pool:
                continue

            poolname = pool.name()
            if poolname not in refresh_pools:
                refresh_pools.append(poolname)

        cloner.start_duplicate(meter)

        for poolname in refresh_pools:
            try:
                pool = self.conn.get_pool_by_name(poolname)
                self.idle_add(pool.refresh)
            except Exception:  # pragma: no cover
                log.debug(
                    "Error looking up pool=%s for refresh after VM clone.",
                    poolname,
                    exc_info=True,
                )

    def _build_final_cloner(self):
        self._set_paths_from_clone_name()

        cloner = self._build_cloner()
        for diskinfo in cloner.get_diskinfos():
            target = diskinfo.disk.target
            sinfo = self._storage_list[target]
            sinfo.set_values_on_diskinfo(diskinfo)

        cloner.prepare()
        for diskinfo in cloner.get_diskinfos():
            diskinfo.raise_error()

        if self._validate(cloner) is False:
            return
        return cloner

    def _finish(self):
        try:
            cloner = self._build_final_cloner()
            if not cloner:
                return
        except Exception as e:
            msg = _("Error with clone settings: %s") % str(e)
            return self.err.show_err(msg)

        self.set_finish_cursor()

        title = _("Creating virtual machine clone '%s'") % cloner.new_guest.name

        text = title
        if cloner.get_nonshare_diskinfos():
            text = (
                _(
                    "Creating virtual machine clone '%s' and selected "
                    "storage (this may take a while)"
                )
                % cloner.new_guest.name
            )

        progWin = vmmAsyncJob(
            self._async_clone,
            [cloner],
            self._finish_cb,
            [self.conn, cloner],
            title,
            text,
            self.topwin,
        )
        progWin.run()

    ################
    # UI listeners #
    ################

    def _close_cb(self, src, event=None):
        return self.close()

    def _finish_clicked_cb(self, src):
        return self._finish()

    def _storage_selection_changed_cb(self, src):
        row = uiutil.get_list_selected_row(self.widget("storage-list"))
        self.widget("storage-details-button").set_sensitive(bool(row))

    def _storage_details_clicked_cb(self, src):
        self._show_storage_window()

    def _storage_dialog_doclone_toggled_cb(self, src):
        do_clone = src.get_active()
        self.widget("change-storage-new").set_sensitive(do_clone)
        self.widget("change-storage-browse").set_sensitive(do_clone)

    def _storage_dialog_close_cb(self, src, event=None):
        return self._storage_dialog_close()

    def _storage_dialog_finish_cb(self, src):
        self._storage_dialog_finish()

    def _storage_dialog_browse_cb(self, ignore):
        def callback(src_ignore, txt):
            self.widget("change-storage-new").set_text(txt)

        if self._storage_browser is None:
            self._storage_browser = vmmStorageBrowser(self.conn)
            self._storage_browser.set_finish_cb(callback)

        self._storage_browser.show(self.topwin)



#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from .device import Device
from ..xmlbuilder import XMLProperty


class DeviceShMem(Device):
    XML_NAME = "shmem"
    _XML_PROP_ORDER = [
        "name",
        "role",
        "type",
        "size",
        "size_unit",
        "server_path",
        "msi_vectors",
        "msi_ioeventfd",
    ]

    MODEL_IVSHMEM = "ivshmem"
    MODEL_IVSHMEM_PLAIN = "ivshmem-plain"
    MODEL_IVSHMEM_DOORBELL = "ivshmem-doorbell"
    MODELS = [MODEL_IVSHMEM, MODEL_IVSHMEM_PLAIN, MODEL_IVSHMEM_DOORBELL]

    ROLE_MASTER = "master"
    ROLE_PEER = "peer"
    ROLES = [ROLE_MASTER, ROLE_PEER]

    name = XMLProperty("./@name")
    role = XMLProperty("./@role")

    type = XMLProperty("./model/@type")

    size = XMLProperty("./size", is_int=True)
    size_unit = XMLProperty("./size/@unit")

    server_path = XMLProperty("./server/@path")
    msi_vectors = XMLProperty("./msi/@vectors", is_int=True)
    msi_ioeventfd = XMLProperty("./msi/@ioeventfd", is_onoff=True)



# Copyright (C) 2007, 2013-2014 Red Hat, Inc.
# Copyright (C) 2007 Daniel P. Berrange <berrange@redhat.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from virtinst import log

from .lib import uiutil
from .baseclass import vmmGObjectUI
from .engine import vmmEngine
from .lib.graphwidgets import Sparkline
from .hostnets import vmmHostNets
from .hoststorage import vmmHostStorage


class vmmHost(vmmGObjectUI):
    @classmethod
    def show_instance(cls, parentobj, conn):
        try:
            # Maintain one dialog per connection
            uri = conn.get_uri()
            if cls._instances is None:
                cls._instances = {}
            if uri not in cls._instances:
                cls._instances[uri] = vmmHost(conn)
            cls._instances[uri].show()
        except Exception as e:  # pragma: no cover
            if not parentobj:
                raise
            parentobj.err.show_err(_("Error launching host dialog: %s") % str(e))

    def __init__(self, conn):
        vmmGObjectUI.__init__(self, "host.ui", "vmm-host")
        self.conn = conn

        # Set default window size
        w, h = self.conn.get_details_window_size()
        if w <= 0:
            w = 800
        if h <= 0:
            h = 600
        self.topwin.set_default_size(w, h)
        self._window_size = None

        self._cpu_usage_graph = None
        self._memory_usage_graph = None
        self._init_conn_state()

        self._storagelist = None
        self._init_storage_state()

        self._hostnets = None
        self._init_net_state()

        self.builder.connect_signals(
            {
                "on_menu_file_view_manager_activate": self._view_manager_cb,
                "on_menu_file_quit_activate": self._exit_app_cb,
                "on_menu_file_close_activate": self.close,
                "on_vmm_host_delete_event": self.close,
                "on_vmm_host_configure_event": self._window_resized_cb,
                "on_host_page_switch": self._page_changed_cb,
                "on_overview_name_changed": self._overview_name_changed_cb,
                "on_config_autoconnect_toggled": self._autoconnect_toggled_cb,
            }
        )

        self.conn.connect("state-changed", self._conn_state_changed_cb)
        self.conn.connect("resources-sampled", self._conn_resources_sampled_cb)

        self._refresh_resources()
        self._refresh_conn_state()
        self.widget("config-autoconnect").set_active(self.conn.get_autoconnect())

        self._cleanup_on_conn_removed()

    #######################
    # Standard UI methods #
    #######################

    def show(self):
        log.debug("Showing host details: %s", self.conn)
        vis = self.is_visible()
        self.topwin.present()
        if vis:
            return  # pragma: no cover

        vmmEngine.get_instance().increment_window_counter()

    def close(self, src=None, event=None):
        dummy = src
        dummy = event
        log.debug("Closing host window for %s", self.conn)
        if not self.is_visible():
            return

        self.topwin.hide()
        vmmEngine.get_instance().decrement_window_counter()

        return 1

    def _cleanup(self):
        if self._window_size:
            self.conn.set_details_window_size(*self._window_size)

        self.conn = None

        self._storagelist.cleanup()
        self._storagelist = None

        self._hostnets.cleanup()
        self._hostnets = None

        self._cpu_usage_graph.destroy()
        self._cpu_usage_graph = None

        self._memory_usage_graph.destroy()
        self._memory_usage_graph = None

    ###########
    # UI init #
    ###########

    def _init_net_state(self):
        self._hostnets = vmmHostNets(self.conn, self.builder, self.topwin)
        self.widget("net-align").add(self._hostnets.top_box)

    def _init_storage_state(self):
        self._storagelist = vmmHostStorage(self.conn, self.builder, self.topwin)
        self.widget("storage-align").add(self._storagelist.top_box)

    def _init_conn_state(self):
        uri = self.conn.get_uri()
        auto = self.conn.get_autoconnect()

        self.widget("overview-uri").set_text(uri)
        self.widget("config-autoconnect").set_active(auto)

        self._cpu_usage_graph = Sparkline()
        self._cpu_usage_graph.set_hexpand(True)
        self._cpu_usage_graph.show()
        self.widget("performance-cpu-align").add(self._cpu_usage_graph)

        self._memory_usage_graph = Sparkline()
        self._memory_usage_graph.set_hexpand(True)
        self._memory_usage_graph.show()
        self.widget("performance-memory-align").add(self._memory_usage_graph)

    ######################
    # UI conn populating #
    ######################

    def _refresh_resources(self):
        vm_memory = uiutil.pretty_mem(self.conn.stats_memory())
        host_memory = uiutil.pretty_mem(self.conn.host_memory_size())

        cpu_vector = self.conn.host_cpu_time_vector()
        memory_vector = self.conn.stats_memory_vector()
        cpu_vector.reverse()
        memory_vector.reverse()

        self.widget("performance-cpu").set_text("%d %%" % self.conn.host_cpu_time_percentage())
        self.widget("performance-memory").set_text(
            _("%(currentmem)s of %(maxmem)s") % {"currentmem": vm_memory, "maxmem": host_memory}
        )

        self._cpu_usage_graph.set_property("data_array", cpu_vector)
        self._memory_usage_graph.set_property("data_array", memory_vector)

    def _refresh_conn_state(self):
        conn_active = self.conn.is_active()

        self.topwin.set_title(
            _("%(connection)s - Connection Details") % {"connection": self.conn.get_pretty_desc()}
        )
        if not self.widget("overview-name").has_focus():
            self.widget("overview-name").set_text(self.conn.get_pretty_desc())

        if conn_active:
            return
        self._hostnets.close()
        self._storagelist.close()

    ################
    # UI listeners #
    ################

    def _view_manager_cb(self, src):
        from .manager import vmmManager

        vmmManager.get_instance(self).show()

    def _exit_app_cb(self, src):
        vmmEngine.get_instance().exit_app()

    def _window_resized_cb(self, src, event):
        if not self.is_visible():
            return
        self._window_size = self.topwin.get_size()

    def _overview_name_changed_cb(self, src):
        src = self.widget("overview-name")
        self.conn.set_config_pretty_name(src.get_text())

    def _autoconnect_toggled_cb(self, src):
        self.conn.set_autoconnect(src.get_active())

    def _page_changed_cb(self, src, child, pagenum):
        if pagenum == 1:
            self._hostnets.refresh_page()
        elif pagenum == 2:
            self._storagelist.refresh_page()

    def _conn_state_changed_cb(self, conn):
        self._refresh_conn_state()

    def _conn_resources_sampled_cb(self, conn):
        self._refresh_resources()



#
# XML API using libxml2
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from . import xmlutil
from .logger import log
from .xmlbase import XMLBase, XPath

# pylint: disable=protected-access


def node_is_text(n):
    return bool(n and n.type == "text")


class Libxml2API(XMLBase):
    def __init__(self, xml):
        XMLBase.__init__(self)

        import libxml2

        self._libxml2 = libxml2

        # Use of gtksourceview in virt-manager changes this libxml
        # global setting which messes up whitespace after parsing.
        # We can probably get away with calling this less but it
        # would take some investigation
        self._libxml2.keepBlanksDefault(1)

        self._doc = self._libxml2.parseDoc(xml)
        self._ctx = self._doc.xpathNewContext()
        self._ctx.setContextNode(self._doc.children)
        for key, val in self.NAMESPACES.items():
            self._ctx.xpathRegisterNs(key, val)

    def __del__(self):
        if not hasattr(self, "_doc"):
            # In case we error when parsing the doc
            return
        self._doc.freeDoc()
        self._doc = None
        self._ctx.xpathFreeContext()
        self._ctx = None

    def _sanitize_xml(self, xml):
        if not xml.endswith("\n") and "\n" in xml:
            xml += "\n"
        return xml

    def copy_api(self):
        return Libxml2API(self._doc.children.serialize())

    def _find(self, fullxpath):
        xpath = XPath(fullxpath).xpath
        try:
            node = self._ctx.xpathEval(xpath)
        except Exception as e:
            log.debug("fullxpath=%s xpath=%s eval failed", fullxpath, xpath, exc_info=True)
            raise RuntimeError("%s %s" % (fullxpath, str(e))) from None
        return node and node[0] or None

    def count(self, xpath):
        return len(self._ctx.xpathEval(xpath))

    def _node_tostring(self, node):
        return node.serialize()

    def _node_from_xml(self, xml):
        return self._libxml2.parseDoc(xml).children

    def _node_get_text(self, node):
        return node.content

    def _node_set_text(self, node, setval):
        if setval is not None:
            setval = xmlutil.xml_escape(setval)
        node.setContent(setval)

    def _node_get_property(self, node, propname):
        prop = node.hasProp(propname)
        if prop:
            return prop.content

    def _node_set_property(self, node, propname, setval):
        if setval is None:
            prop = node.hasProp(propname)
            if prop:
                prop.unlinkNode()
                prop.freeNode()
        else:
            node.setProp(propname, setval)

    def _node_new(self, xpathseg, parentnode):
        newnode = self._libxml2.newNode(xpathseg.nodename)
        if not xpathseg.nsname:
            return newnode

        def _find_parent_ns():
            parent = parentnode
            while parent:
                for ns in xmlutil.listify(parent.nsDefs()):
                    if ns.name == xpathseg.nsname:
                        return ns
                parent = parent.get_parent()

        ns = _find_parent_ns()
        if not ns:
            ns = newnode.newNs(self.NAMESPACES[xpathseg.nsname], xpathseg.nsname)
        newnode.setNs(ns)
        return newnode

    def node_clear(self, xpath):
        node = self._find(xpath)
        if node:
            propnames = [p.name for p in (node.properties or [])]
            for p in propnames:
                node.unsetProp(p)
            node.setContent(None)

    def _node_has_content(self, node):
        return node.type == "element" and (node.children or node.properties)

    def _node_get_name(self, node):
        return node.name

    def _node_remove_child(self, parentnode, childnode):
        node = childnode

        # Look for preceding whitespace and remove it
        white = node.get_prev()
        if node_is_text(white):
            white.unlinkNode()
            white.freeNode()

        node.unlinkNode()
        node.freeNode()
        if all([node_is_text(n) for n in parentnode.children]):
            parentnode.setContent(None)

    def _node_add_child(self, parentxpath, parentnode, newnode):
        ignore = parentxpath
        if not node_is_text(parentnode.get_last()):
            prevsib = parentnode.get_prev()
            if node_is_text(prevsib):
                newlast = self._libxml2.newText(prevsib.content)
            else:
                newlast = self._libxml2.newText("\n")
            parentnode.addChild(newlast)

        endtext = parentnode.get_last().content
        parentnode.addChild(self._libxml2.newText("  "))
        parentnode.addChild(newnode)
        parentnode.addChild(self._libxml2.newText(endtext))

    def _node_replace_child(self, xpath, newnode):
        oldnode = self._find(xpath)
        oldnode.replaceNode(newnode)



#
# Copyright 2013 Fujitsu Limited.
# Chen Hanxiao <chenhanxiao at cn.fujitsu.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from .device import Device
from ..xmlbuilder import XMLProperty


class DevicePanic(Device):
    XML_NAME = "panic"

    model = XMLProperty("./@model")
    set_stub = XMLProperty(".", is_bool=True)

    ##################
    # Default config #
    ##################

    def set_defaults(self, guest):
        if not self.address.type and self.address.iobase:
            self.address.type = "isa"
        if not self.model:
            # This asks libvirt to fill in a default
            self.set_stub = True



#
# Copyright 2013, 2015 Red Hat, Inc.
# Copyright(c) FUJITSU Limited 2007.
#
# Cloning a virtual machine module.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import re
import os
from itertools import chain

import libvirt

from . import generatename
from . import progress
from . import xmlutil
from .guest import Guest
from .devices import DeviceInterface
from .devices import DeviceDisk
from .logger import log
from .devices import DeviceChannel
from .devices import DeviceSerial


def _replace_vm(conn, name):
    """
    Remove the existing VM with the same name if requested
    """
    try:
        vm = conn.lookupByName(name)
    except libvirt.libvirtError:
        return

    try:

        log.debug("Explicitly replacing guest '%s'", name)
        if vm.ID() != -1:
            log.debug("Destroying guest '%s'", name)
            vm.destroy()

        log.debug("Undefining guest '%s'", name)
        vm.undefine()
    except libvirt.libvirtError as e:  # pragma: no cover
        msg = _("Could not remove old vm '%(vm)s': %(error)s") % {"vm": name, "error": str(e)}
        raise RuntimeError(msg) from None


def _generate_clone_name(conn, basename):
    """
    If the orig name is "foo-clone", we don't want the clone to be
    "foo-clone-clone", we want "foo-clone1"
    """
    regex = r"-clone[1-9]*$"
    match = re.search(regex, basename)
    start_num = 1
    force_num = False
    if match:
        num_match = re.search("[1-9]+$", match.group())
        force_num = True
        if num_match:
            start_num = int(str(num_match.group())) + 1
        basename = basename[: match.start()]

    def cb(n):
        return generatename.check_libvirt_collision(conn.lookupByName, n)

    basename = basename + "-clone"
    return generatename.generate_name(
        basename, cb, sep="", start_num=start_num, force_num=force_num
    )


def _generate_clone_path(origname, newname, origpath, cb_exists):
    """
    Generate desired cloned path for auxiliary files, derived from the
    original path, original VM name, and proposed new VM name
    """
    if origpath is None:
        return None

    path = origpath
    suffix = ""

    # Try to split the suffix off the existing disk name. Ex.
    # foobar.img -> foobar-clone.img
    #
    # If the suffix is greater than 7 characters, assume it isn't
    # a file extension and is part of the disk name, at which point
    # just stick '-clone' on the end.
    if "." in origpath and len(origpath.rsplit(".", 1)[1]) <= 7:
        path, suffix = origpath.rsplit(".", 1)
        suffix = "." + suffix

    dirname = os.path.dirname(path)
    basename = os.path.basename(path)

    clonebase = basename + "-clone"
    if origname and basename == origname:
        clonebase = newname

    clonebase = os.path.join(dirname, clonebase)
    return generatename.generate_name(clonebase, cb_exists, suffix=suffix)


def _lookup_vm(conn, name):
    try:
        return conn.lookupByName(name)
    except libvirt.libvirtError:
        e = ValueError(_("Domain '%s' was not found.") % str(name))
        raise e from None


def _build_clone_vol_install(orig_disk, new_disk):
    # We set a stub size for initial creation
    # set_input_vol will overwrite it
    size = 0.000001
    sparse = False
    vol_install = DeviceDisk.build_vol_install(
        orig_disk.conn,
        os.path.basename(new_disk.get_source_path()),
        new_disk.get_parent_pool(),
        size,
        sparse,
    )
    vol_install.set_input_vol(orig_disk.get_vol_object())

    return vol_install


def _build_clone_disk(orig_disk, clonepath, allow_create, sparse):
    conn = orig_disk.conn
    device = DeviceDisk.DEVICE_DISK
    if not clonepath:
        device = DeviceDisk.DEVICE_CDROM

    new_disk = DeviceDisk(conn)
    new_disk.set_source_path(clonepath)
    new_disk.device = device

    if not allow_create:
        new_disk.validate()
        return new_disk

    if new_disk.get_vol_object():
        # Special case: non remote cloning of a guest using
        # managed block devices: fall back to local cloning if
        # we have permissions to do so. This validation check
        # caused a few bug reports in a short period of time,
        # so must be a common case.
        if (
            conn.is_remote()
            or new_disk.type != new_disk.TYPE_BLOCK
            or not orig_disk.get_source_path()
            or not os.access(orig_disk.get_source_path(), os.R_OK)
            or not new_disk.get_source_path()
            or not os.access(new_disk.get_source_path(), os.W_OK)
        ):
            raise RuntimeError(
                _("Clone onto existing storage volume is not currently supported: '%s'")
                % new_disk.get_source_path()
            )

    if orig_disk.get_vol_object() and new_disk.wants_storage_creation():
        vol_install = _build_clone_vol_install(orig_disk, new_disk)
        if not sparse:
            vol_install.allocation = vol_install.capacity
        new_disk.set_vol_install(vol_install)
    elif orig_disk.get_source_path():
        new_disk.set_local_disk_to_clone(orig_disk, sparse)

    new_disk.validate()
    return new_disk


def _get_cloneable_msg(disk):
    """
    If the disk storage is not cloneable, return a string explaining why
    """
    if disk.wants_storage_creation():
        return _("Disk path '%s' does not exist.") % disk.get_source_path()

    if disk.type == "network":
        proto = disk.source.protocol
        if proto == "rbd":
            # This case, rbd with managed storage, is implementable. It
            # requires open coding a bunch of work in cloner, or reworking
            # other disk code to add unique URIs for rbd volumes and pools
            return (
                _("Cloning rbd volumes is not yet supported.")
                + " https://github.com/virt-manager/virt-manager/issues/177"
            )
        return _("Disk network type '%s' is not cloneable.") % proto


def _get_shareable_msg(disk):
    if disk.is_empty():
        return _("No storage to clone.")
    if disk.read_only:
        return _("Read Only")
    if disk.shareable or disk.transient_shareBacking:
        return _("Marked as shareable")


class _CloneDiskInfo:
    """
    Class that tracks some additional information about how we want
    to default handle each disk of the source VM

    For any source disk there's 3 main scenarios:

    * clone: Copy contents from src to dst. If dst path doesn't
            exist we attempt to create it. If it exists we overwrite it
    * preserve: Destination path is an existing, and no copying is performed.
    * share: Original disk XML is used unchanged for the new disk
    """

    _ACTION_SHARE = 1
    _ACTION_CLONE = 2
    _ACTION_PRESERVE = 3

    def __init__(self, srcdisk):
        self.disk = DeviceDisk(srcdisk.conn, parsexml=srcdisk.get_xml())
        self.new_disk = None

        self._cloneable_msg = -1
        try:
            # Failure here means source path may not exist. Sometimes
            # that's fatal, sometimes it's not (like with `--preserve`),
            # but we don't know for sure until later.
            self.disk.set_backend_for_existing_path()
        except Exception as e:
            self._cloneable_msg = str(e)

        self._share_msg = _get_shareable_msg(self.disk)
        self._newpath_msg = None

        self._action = None
        self.set_clone_requested()
        if self.get_share_msg():
            self.set_share_requested()

    def is_clone_requested(self):
        return self._action in [self._ACTION_CLONE]

    def is_share_requested(self):
        return self._action in [self._ACTION_SHARE]

    def is_preserve_requested(self):
        return self._action in [self._ACTION_PRESERVE]

    def _set_action(self, action):
        if action != self._action:
            self._action = action

    def set_clone_requested(self):
        self._set_action(self._ACTION_CLONE)

    def set_share_requested(self):
        self._set_action(self._ACTION_SHARE)

    def set_preserve_requested(self):
        self._set_action(self._ACTION_PRESERVE)

    def set_new_path(self, path, sparse):
        allow_create = not self.is_preserve_requested()
        if allow_create:
            msg = self.get_cloneable_msg()
            if msg:
                return

        try:
            self.new_disk = Cloner.build_clone_disk(self.disk, path, allow_create, sparse)
        except Exception as e:
            log.debug("Error setting clone path.", exc_info=True)
            err = _("Could not use path '%(path)s' for cloning: %(error)s") % {
                "path": path,
                "error": str(e),
            }
            self._newpath_msg = err

    def get_share_msg(self):
        return self._share_msg

    def get_cloneable_msg(self):
        if self._cloneable_msg == -1:
            self._cloneable_msg = _get_cloneable_msg(self.disk)
        return self._cloneable_msg

    def get_newpath_msg(self):
        return self._newpath_msg

    def raise_error(self):
        if self.is_clone_requested() and self.get_cloneable_msg():
            msg = self.get_cloneable_msg()
            err = _("Could not determine original disk information: %s") % msg
            raise ValueError(err)
        if self.is_share_requested():
            return
        if self.get_newpath_msg():
            msg = self.get_newpath_msg()
            raise ValueError(msg)


class Cloner:
    @staticmethod
    def generate_clone_name(conn, basename):
        return _generate_clone_name(conn, basename)

    @staticmethod
    def generate_clone_disk_path(conn, origname, newname, origpath):
        def cb_exists(p):
            return DeviceDisk.path_definitely_exists(conn, p)

        return _generate_clone_path(origname, newname, origpath, cb_exists)

    @staticmethod
    def build_clone_disk(orig_disk, clonepath, allow_create, sparse):
        return _build_clone_disk(orig_disk, clonepath, allow_create, sparse)

    def __init__(self, conn, src_name=None, src_xml=None):
        self.conn = conn

        self._src_guest = None
        self._new_guest = None
        self._diskinfos = []
        self._nvram_diskinfo = None
        self._init_src(src_name, src_xml)

        self._new_nvram_path = None

        self._sparse = True
        self._replace = False
        self._reflink = False

    #################
    # Init routines #
    #################

    def _init_src(self, src_name, src_xml):
        """
        Set up the source VM info we are cloning, from passed in VM name
        or full XML
        """
        if not src_xml:
            dom = _lookup_vm(self.conn, src_name)
            status = dom.info()[0]
            if status not in [libvirt.VIR_DOMAIN_SHUTOFF]:
                raise RuntimeError(_("Domain to clone must be shutoff."))
            flags = libvirt.VIR_DOMAIN_XML_SECURE
            src_xml = dom.XMLDesc(flags)

        log.debug("Original XML:\n%s", src_xml)

        self._src_guest = Guest(self.conn, parsexml=src_xml)
        self._new_guest = Guest(self.conn, parsexml=src_xml)
        self._init_new_guest()

        # Collect disk info for every disk to determine if we will
        # default to cloning or not
        for disk in self._src_guest.devices.disk:
            self._diskinfos.append(_CloneDiskInfo(disk))
        for diskinfo in [d for d in self._diskinfos if d.is_clone_requested()]:
            disk = diskinfo.disk
            log.debug("Wants cloning: size=%s path=%s", disk.get_size(), disk.get_source_path())

        if self._src_guest.os.nvram:
            old_nvram = DeviceDisk(self.conn)
            old_nvram.set_source_path(self._new_guest.os.nvram)
            self._nvram_diskinfo = _CloneDiskInfo(old_nvram)

    def _init_new_guest(self):
        """
        Perform the series of unconditional new VM changes we always make
        """
        self._new_guest.id = None
        self._new_guest.title = None
        self.set_clone_uuid(Guest.generate_uuid(self.conn))

        for dev in self._new_guest.devices.graphics:
            if dev.port and dev.port != -1:
                log.warning(
                    _(
                        "Setting the graphics device port to autoport, "
                        "in order to avoid conflicting."
                    )
                )
                dev.port = -1

        for iface in self._new_guest.devices.interface:
            iface.target_dev = None
            iface.macaddr = DeviceInterface.generate_mac(self.conn)

        # For guest agent channel, remove a path to generate a new one with
        # new guest name
        for channel in self._new_guest.devices.channel:
            if (
                channel.type == DeviceChannel.TYPE_UNIX
                and channel.target_name
                and channel.source.path
                and channel.target_name in channel.source.path
            ):
                channel.source.path = None

        new_name = Cloner.generate_clone_name(self.conn, self.src_name)
        log.debug("Auto-generated clone name '%s'", new_name)
        self.set_clone_name(new_name)

    ##############
    # Properties #
    ##############

    @property
    def src_name(self):
        """
        The name of the original VM we are cloning
        """
        return self._src_guest.name

    @property
    def new_guest(self):
        """
        The Guest instance of the new XML we will create
        """
        return self._new_guest

    @property
    def nvram_diskinfo(self):
        return self._nvram_diskinfo

    def set_clone_name(self, name):
        self._new_guest.name = name

    def set_clone_uuid(self, uuid):
        """
        Override the new VMs generated UUId
        """
        self._new_guest.uuid = uuid
        for sysinfo in self._new_guest.sysinfo:
            if sysinfo.system_uuid:
                sysinfo.system_uuid = uuid

    def set_replace(self, val):
        """
        If True, don't check for clone name collision, simply undefine
        any conflicting guest.
        """
        self._replace = bool(val)

    def set_reflink(self, reflink):
        """
        If true, use COW lightweight copy
        """
        self._reflink = reflink

    def set_sparse(self, flg):
        """
        If True, attempt sparse allocation during cloning
        """
        self._sparse = flg

    def get_diskinfos(self):
        """
        Return the list of _CloneDiskInfo instances
        """
        return self._diskinfos[:]

    def get_nonshare_diskinfos(self):
        """
        Return a list of _CloneDiskInfo that are tagged for cloning
        """
        return [di for di in self.get_diskinfos() if not di.is_share_requested()]

    def set_nvram_path(self, val):
        """
        If the VM needs to have nvram content cloned, this overrides the
        destination path
        """
        self._new_nvram_path = val

    ######################
    # Functional methods #
    ######################

    def _prepare_serial_files(self):
        for serial in chain(self._new_guest.devices.console, self._new_guest.devices.serial):
            if serial.type != DeviceSerial.TYPE_FILE:
                continue

            def cb_exists(path):
                # Ignore the check for now
                return False

            serial.source.path = _generate_clone_path(
                self.src_name, self.new_guest.name, serial.source.path, cb_exists=cb_exists
            )

    def _prepare_nvram(self):
        if not self._nvram_diskinfo:
            return

        diskinfo = self._nvram_diskinfo
        old_nvram = DeviceDisk(self.conn)
        old_nvram_path = diskinfo.disk.get_source_path()
        old_nvram.set_source_path(old_nvram_path)

        new_nvram_path = self._new_nvram_path
        if new_nvram_path is None:
            ext = os.path.splitext(old_nvram_path)[1]
            nvram_dir = os.path.dirname(old_nvram_path)
            new_nvram_path = os.path.join(
                nvram_dir, "%s_VARS%s" % (os.path.basename(self._new_guest.name), ext or ".fd")
            )

        new_nvram = DeviceDisk(self.conn)
        new_nvram.set_source_path(new_nvram_path)

        if (
            diskinfo.is_clone_requested()
            and new_nvram.wants_storage_creation()
            and diskinfo.disk.get_vol_object()
        ):
            # We only run validation if there's some existing nvram we
            # can copy. It's valid for nvram to not exist at VM define
            # time, libvirt will create it for us
            diskinfo.set_new_path(new_nvram_path, self._sparse)
            diskinfo.raise_error()
            diskinfo.new_disk.get_vol_install().reflink = self._reflink
        else:
            # There's no action to perform for this case, so drop it
            self._nvram_diskinfo = None

        self._new_guest.os.nvram = new_nvram.get_source_path()

    def prepare(self):
        """
        Validate and set up all parameters needed for the new (clone) VM
        """
        try:
            Guest.validate_name(
                self.conn, self._new_guest.name, check_collision=not self._replace, validate=False
            )
        except ValueError as e:
            msg = _("Invalid name for new guest: %s") % e
            raise ValueError(msg) from None

        for diskinfo in self.get_nonshare_diskinfos():
            orig_disk = diskinfo.disk

            if not diskinfo.new_disk:
                # User didn't set a path, generate one
                newpath = Cloner.generate_clone_disk_path(
                    self.conn, self.src_name, self.new_guest.name, orig_disk.get_source_path()
                )
                diskinfo.set_new_path(newpath, self._sparse)
                if not diskinfo.new_disk:
                    # We hit an error, clients will raise it later
                    continue

            new_disk = diskinfo.new_disk
            assert new_disk
            log.debug(
                "Cloning srcpath=%s dstpath=%s",
                orig_disk.get_source_path(),
                new_disk.get_source_path(),
            )

            if self._reflink:
                vol_install = new_disk.get_vol_install()
                vol_install.reflink = self._reflink

            for disk in self._new_guest.devices.disk:
                if disk.target == orig_disk.target:
                    xmldisk = disk

            # Change the XML
            xmldisk.set_source_path(None)
            xmldisk.type = new_disk.type
            xmldisk.driver_name = orig_disk.driver_name
            xmldisk.driver_type = orig_disk.driver_type
            xmldisk.set_source_path(new_disk.get_source_path())

        self._prepare_nvram()
        self._prepare_serial_files()

        # Save altered clone xml
        diff = xmlutil.diff(self._src_guest.get_xml(), self._new_guest.get_xml())
        log.debug("Clone guest xml diff:\n%s", diff)

    def start_duplicate(self, meter=None):
        """
        Actually perform the duplication: cloning disks if needed and defining
        the new clone xml.
        """
        log.debug("Starting duplicate.")
        meter = progress.ensure_meter(meter)

        dom = None
        try:
            # Replace orig VM if required
            if self._replace:
                _replace_vm(self.conn, self._new_guest.name)

            # Define domain early to catch any xml errors before duping storage
            dom = self.conn.defineXML(self._new_guest.get_xml())

            diskinfos = self.get_diskinfos()
            if self._nvram_diskinfo:
                diskinfos.append(self._nvram_diskinfo)

            for diskinfo in diskinfos:
                if not diskinfo.is_clone_requested():
                    continue
                diskinfo.new_disk.build_storage(meter)
        except Exception as e:
            log.debug("Duplicate failed: %s", str(e))
            if dom:
                dom.undefine()
            raise

        log.debug("Duplicating finished.")



#
# Some code for parsing libvirt's capabilities XML
#
# Copyright 2007, 2012-2014 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import pwd

from .logger import log
from .xmlbuilder import XMLBuilder, XMLChildProperty, XMLProperty


###################################
# capabilities host <cpu> parsing #
###################################


class _CapsCPU(XMLBuilder):
    XML_NAME = "cpu"
    arch = XMLProperty("./arch")
    model = XMLProperty("./model")
    vendor = XMLProperty("./vendor")


######################################
# Caps <host> and <secmodel> parsers #
######################################


class _CapsSecmodelBaselabel(XMLBuilder):
    XML_NAME = "baselabel"
    type = XMLProperty("./@type")
    content = XMLProperty(".")


class _CapsSecmodel(XMLBuilder):
    XML_NAME = "secmodel"
    model = XMLProperty("./model")
    baselabels = XMLChildProperty(_CapsSecmodelBaselabel)


class _CapsHost(XMLBuilder):
    XML_NAME = "host"
    secmodels = XMLChildProperty(_CapsSecmodel)
    cpu = XMLChildProperty(_CapsCPU, is_single=True)

    def get_qemu_baselabel(self):
        for secmodel in self.secmodels:
            if secmodel.model != "dac":
                continue

            label = None
            for baselabel in secmodel.baselabels:
                if baselabel.type in ["qemu", "kvm"]:
                    label = baselabel.content
                    break
            if not label:
                continue  # pragma: no cover

            # XML we are looking at is like:
            #
            # <secmodel>
            #   <model>dac</model>
            #   <doi>0</doi>
            #   <baselabel type='kvm'>+107:+107</baselabel>
            #   <baselabel type='qemu'>+107:+107</baselabel>
            # </secmodel>
            try:
                uid = int(label.split(":")[0].replace("+", ""))
                user = pwd.getpwuid(uid)[0]
                return user, uid
            except Exception:
                log.debug("Exception parsing qemu dac baselabel=%s", label, exc_info=True)
        return None, None


################################
# <guest> and <domain> parsers #
################################


class _CapsMachine(XMLBuilder):
    XML_NAME = "machine"
    name = XMLProperty(".")
    canonical = XMLProperty("./@canonical")


class _CapsDomain(XMLBuilder):
    XML_NAME = "domain"
    hypervisor_type = XMLProperty("./@type")
    emulator = XMLProperty("./emulator")
    machines = XMLChildProperty(_CapsMachine)


class _CapsGuestFeatures(XMLBuilder):
    XML_NAME = "features"

    pae = XMLProperty("./pae", is_bool=True)
    acpi = XMLProperty("./acpi/@default", is_onoff=True)
    apic = XMLProperty("./apic/@default", is_onoff=True)
    externalSnapshot = XMLProperty("./externalSnapshot", is_bool=True)


class _CapsGuest(XMLBuilder):
    XML_NAME = "guest"

    os_type = XMLProperty("./os_type")
    arch = XMLProperty("./arch/@name")
    loader = XMLProperty("./arch/loader")
    emulator = XMLProperty("./arch/emulator")

    domains = XMLChildProperty(_CapsDomain, relative_xpath="./arch")
    features = XMLChildProperty(_CapsGuestFeatures, is_single=True)
    machines = XMLChildProperty(_CapsMachine, relative_xpath="./arch")

    ###############
    # Public APIs #
    ###############

    def all_machine_names(self, domain):
        """
        Return all machine string names, including canonical aliases for
        the guest+domain combo but avoiding duplicates
        """
        mobjs = (domain and domain.machines) or self.machines
        ret = []
        for m in mobjs:
            ret.append(m.name)
            if m.canonical and m.canonical not in ret:
                ret.append(m.canonical)  # pragma: no cover
        return ret

    def is_machine_alias(self, domain, src, tgt):
        """
        Determine if machine @src is an alias for machine @tgt
        """
        mobjs = (domain and domain.machines) or self.machines
        for m in mobjs:
            if m.name == src and m.canonical == tgt:
                return True
        return False

    def is_kvm_available(self):
        """
        Return True if kvm guests can be installed
        """
        for d in self.domains:
            if d.hypervisor_type == "kvm":
                return True
        return False

    def supports_pae(self):
        """
        Return True if capabilities report support for PAE
        """
        return bool(self.features.pae)

    def supports_acpi(self):
        """
        Return Tree if capabilities report support for ACPI
        """
        return bool(self.features.acpi)

    def supports_apic(self):
        """
        Return Tree if capabilities report support for APIC
        """
        return bool(self.features.apic)

    def supports_externalSnapshot(self):
        """
        Return True if capabilities report support for external snapshots
        """
        return bool(self.features.externalSnapshot)


############################
# Main capabilities object #
############################


class _CapsInfo:
    """
    Container object to hold the results of guest_lookup, so users don't
    need to juggle two objects
    """

    def __init__(self, conn, guest, domain):
        self.conn = conn
        self.guest = guest
        self.domain = domain

        self.hypervisor_type = self.domain.hypervisor_type
        self.os_type = self.guest.os_type
        self.arch = self.guest.arch
        self.loader = self.guest.loader

        self.emulator = self.domain.emulator or self.guest.emulator
        self.machines = self.guest.all_machine_names(self.domain)

    def is_machine_alias(self, src, tgt):
        return self.guest.is_machine_alias(self.domain, src, tgt)


class Capabilities(XMLBuilder):
    def __init__(self, *args, **kwargs):
        XMLBuilder.__init__(self, *args, **kwargs)
        self._cpu_models_cache = {}

    XML_NAME = "capabilities"

    host = XMLChildProperty(_CapsHost, is_single=True)
    guests = XMLChildProperty(_CapsGuest)

    ############################
    # Public XML building APIs #
    ############################

    def _guestForOSType(self, os_type, arch):
        archs = [arch]
        if arch is None:
            archs = [self.host.cpu.arch, None]

        for a in archs:
            for g in self.guests:
                if (os_type is None or g.os_type == os_type) and (a is None or g.arch == a):
                    return g

    def _bestDomainType(self, guest, dtype, machine):
        """
        Return the recommended domain for use if the user does not explicitly
        request one.
        """
        domains = []
        for d in guest.domains:
            if dtype and d.hypervisor_type != dtype.lower():
                continue
            if machine and machine not in guest.all_machine_names(d):
                continue

            domains.append(d)

        if not domains:
            return None

        priority = ["kvm", "xen", "hvf", "qemu"]

        for t in priority:
            for d in domains:
                if d.hypervisor_type == t:
                    return d

        # Fallback, just return last item in list
        return domains[-1]

    def has_install_options(self):
        """
        Return True if there are any install options available
        """
        for guest in self.guests:
            if guest.domains:
                return True
        return False

    def guest_lookup(self, os_type=None, arch=None, typ=None, machine=None):
        """
        Simple virtualization availability lookup

        Convenience function for looking up 'Guest' and 'Domain' capabilities
        objects for the desired virt type. If type, arch, or os_type are none,
        we return the default virt type associated with those values. These are
        typically:

            - os_type : hvm, then xen
            - typ     : kvm over plain qemu
            - arch    : host arch over all others

        Otherwise the default will be the first listed in the capabilities xml.
        This function throws C{ValueError}s if any of the requested values are
        not found.

        :param typ: Virtualization type ('hvm', 'xen', ...)
        :param arch: Guest architecture ('x86_64', 'i686' ...)
        :param os_type: Hypervisor name ('qemu', 'kvm', 'xen', ...)
        :param machine: Optional machine type to emulate

        :returns: A _CapsInfo object containing the found guest and domain
        """
        # F22 libxl xen still puts type=linux in the XML, so we need
        # to handle it for caps lookup
        if os_type == "linux":
            os_type = "xen"

        guest = self._guestForOSType(os_type, arch)
        if not guest:
            if arch and os_type:
                msg = _(
                    "Host does not support virtualization type "
                    "'%(virttype)s' for architecture '%(arch)s'"
                ) % {"virttype": os_type, "arch": arch}
            elif arch:
                msg = _(
                    "Host does not support any virtualization options "
                    "for architecture '%(arch)s'"
                ) % {"arch": arch}
            elif os_type:
                msg = _("Host does not support virtualization type '%(virttype)s'") % {
                    "virttype": os_type
                }
            else:
                msg = _("Host does not support any virtualization options")
            raise ValueError(msg)

        domain = self._bestDomainType(guest, typ, machine)
        if domain is None:
            if machine:
                msg = _(
                    "Host does not support domain type %(domain)s with "
                    "machine '%(machine)s' for virtualization type "
                    "'%(virttype)s' with architecture '%(arch)s'"
                ) % {
                    "domain": typ,
                    "virttype": guest.os_type,
                    "arch": guest.arch,
                    "machine": machine,
                }
            else:
                msg = _(
                    "Host does not support domain type %(domain)s for "
                    "virtualization type '%(virttype)s' with "
                    "architecture '%(arch)s'"
                ) % {"domain": typ, "virttype": guest.os_type, "arch": guest.arch}
            raise ValueError(msg)

        capsinfo = _CapsInfo(self.conn, guest, domain)
        return capsinfo



#
# XML API using stock python ElementTree
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import io
import re
import lxml.etree as lxmletree
import xml.etree.ElementTree as ET

from . import xmlutil
from .xmlbase import XMLBase, XPath

# We need to extend ElementTree to parse + rebuild XML with no diff
# from default libvirt output. Otherwise `virt-xml --edit` diffs
# are needlessly noisy.
#
# The main problematic area is xmlns namespace handling.
#
# 1) libvirt xml will preserve arbitrary xml definitions.
#    ElementTree will _rename_ xmlns definition to ns0, ns1, etc
#    unless `register_namespace` was called ahead of time.
#
# 2) ElementTree formats every xmlns attribute into the top
#    element of the document, but libvirt may keep them inline,
#    like for <domain> <metadata>.


class _VirtinstElement(ET.Element):
    """
    Wrap Element to track specifically where an xmlns
    was defined. Default ElementTree throws this away
    """

    def __init__(self, *args, **kwargs):
        self.virtinst_namespaces = {}
        ET.Element.__init__(self, *args, **kwargs)

    def virtinst_add_namespace(self, prefix, uri):
        self.virtinst_namespaces[prefix] = uri


def _fromstring(xml):
    namespaces = {}

    class _VirtinstTreeBuilder(ET.TreeBuilder):
        """
        Custom tree builder to do two things:

        1) track element where xmlns attribute was defined
        2) build a mapping of xmlns prefix:uri for every xmlns we see
        """

        _ns_stack = []
        _last_element = None

        def end(self, tag):
            self._last_element = ET.TreeBuilder.end(self, tag)
            return self._last_element

        def start_ns(self, prefix, uri):
            self._ns_stack.append((prefix, uri))
            return (prefix, uri)

        def end_ns(self, _prefix):
            prefix, uri = self._ns_stack.pop()
            self._last_element.virtinst_add_namespace(prefix, uri)
            namespaces[prefix] = uri
            return prefix

    builder = _VirtinstTreeBuilder(element_factory=_VirtinstElement, insert_comments=True)
    parser = ET.XMLParser(target=builder)
    parser.feed(xml)
    node = parser.close()
    return node, namespaces


def _escape_cdata(xml):
    if xml:
        xml = xml.replace("&", "&amp;")
        xml = xml.replace("<", "&lt;")
        xml = xml.replace(">", "&gt;")
    return xml


def _convert_qname(tag, namespaces):
    """
    Convert ElementTree style namespace names to final
    XML format. For example, given this XML:

    <MYNS:FOO xmlns:MYNS="http://example.com"/>

    ElementTree node.tag will be "{http://example.com}FOO",
    and we turn it back into "MYNS:FOO"
    """
    if tag and tag.startswith("{"):
        uri, tag = tag[1:].rsplit("}", 1)
        for key, val in namespaces.items():
            if uri == val:
                tag = key + ":" + tag
                break
    return tag


def _serialize_node(write, elem, namespaces):
    # derived from ElementTree._serialize_xml
    tag = elem.tag
    text = elem.text
    if tag is ET.Comment:
        write("<!--%s-->" % text)
    else:
        use_ns = elem.virtinst_namespaces.copy()
        use_ns.update(namespaces)

        tag = _convert_qname(tag, use_ns)

        if tag is None:  # pragma: no cover
            # This is for CDATA, which libvirt will throw out anyways.
            pass
        else:
            write("<" + tag)
            for nsprefix, nsuri in elem.virtinst_namespaces.items():
                write(' xmlns:%s="%s"' % (nsprefix, nsuri))
            for k, v in list(elem.items()):
                k = _convert_qname(k, use_ns)
                v = xmlutil.xml_escape(v)
                write(' %s="%s"' % (k, v))

            if text or len(elem):
                write(">")
                if text:
                    write(_escape_cdata(text))
                for e in elem:
                    _serialize_node(write, e, namespaces)
                write("</" + tag + ">")
            else:
                write("/>")

    if elem.tail:
        write(_escape_cdata(elem.tail))


def _tostring(node, namespaces):
    stream = io.StringIO()

    _serialize_node(stream.write, node, namespaces)
    ret = stream.getvalue()
    return ret.rstrip()


class ETreeAPI(XMLBase):
    def __init__(self, parsexml):
        XMLBase.__init__(self)
        node, namespaces = _fromstring(parsexml)
        self._et = ET.ElementTree(node)
        self._namespaces = namespaces

    #######################
    # Private helper APIs #
    #######################

    def _sanitize_xml(self, xml):
        return xml

    def _node_tostring(self, node):
        return _tostring(node, self._namespaces)

    def _node_from_xml(self, xml):
        return _fromstring(xml)[0]

    def _node_get_name(self, node):
        name = _convert_qname(node.tag, self._namespaces)
        if ":" in name:
            name = name.split(":", 1)[1]
        return name

    def _node_get_text(self, node):
        return node.text

    def _node_set_text(self, node, setval):
        node.text = setval

    def _node_get_property(self, node, propname):
        return node.attrib.get(propname)

    def _node_set_property(self, node, propname, setval):
        if setval is None:
            node.attrib.pop(propname, None)
        else:
            node.attrib[propname] = setval

    def _find(self, fullxpath):
        xpath = XPath(fullxpath).xpath

        root = "/" + self._node_get_name(self._et.getroot())
        if xpath.startswith(root):
            # ElementTree explicitly warns that absolute xpaths don't
            # work as expected, and need a prepended .
            xpath = "." + xpath[len(root) :]

        node = self._et.find(xpath, self.NAMESPACES)
        if node is None:
            return None
        return node

    ###############
    # Simple APIs #
    ###############

    def copy_api(self):
        return ETreeAPI(self._node_tostring(self._et.getroot()))

    def count(self, xpath):
        return len(self._et.findall(xpath, self.NAMESPACES) or [])

    ####################
    # Private XML APIs #
    ####################

    def _node_add_child(self, parentxpath, parentnode, newnode):
        """
        Add 'newnode' as a child of 'parentnode', but try to preserve
        whitespace and nicely format the result.
        """
        xpathobj = XPath(parentxpath)

        if bool(len(parentnode)):
            lastelem = list(parentnode)[-1]
            newnode.tail = lastelem.tail
            lastelem.tail = parentnode.text
        elif xpathobj.parent_xpath():
            grandparent = self._find(xpathobj.parent_xpath())
            idx = list(grandparent).index(parentnode)
            if idx == (len(list(grandparent)) - 1):
                parentnode.text = (grandparent.text or "\n") + "  "
                newnode.tail = (parentnode.tail or "\n") + "  "
            else:
                parentnode.text = list(grandparent)[0].tail + "  "
                newnode.tail = list(grandparent)[0].tail
        else:
            parentnode.text = "\n  "
            newnode.tail = "\n"

        parentnode.append(newnode)

    def _node_has_content(self, node):
        return len(node) or node.attrib or re.search(r"\w+", (node.text or ""))

    def _node_remove_child(self, parentnode, childnode):
        idx = list(parentnode).index(childnode)

        if idx != 0 and idx == (len(list(parentnode)) - 1):
            prevsibling = list(parentnode)[idx - 1]
            prevsibling.tail = prevsibling.tail[:-2]
        elif idx == 0 and len(list(parentnode)) == 1:
            parentnode.text = None

        parentnode.remove(childnode)

    def _node_new(self, xpathseg, _parentnode):
        newname = xpathseg.nodename
        nsname = xpathseg.nsname
        nsuri = self.NAMESPACES.get(nsname, None)

        if nsname:
            newname = "{%s}%s" % (nsuri, newname)
        element = _VirtinstElement(newname)
        if nsname and nsname not in self._namespaces:
            self._namespaces[nsname] = nsuri
            element.virtinst_add_namespace(nsname, nsuri)
        return element

    def _node_replace_child(self, xpath, newnode):
        oldnode = self._find(xpath)
        parentnode = self._find(xpath + "...")
        for idx, elem in list(enumerate(parentnode)):
            if elem != oldnode:
                continue
            newnode.tail = oldnode.tail
            parentnode.remove(oldnode)
            parentnode.insert(idx, newnode)
            break

    ####################
    # XML editing APIs #
    ####################

    def node_clear(self, xpath):
        node = self._find(xpath)
        if node is not None:
            for c in list(node):
                node.remove(c)
            node.attrib.clear()
            node.text = None
